import numpy as np
import pytest

import starfish.data
from starfish import FieldOfView
from starfish.image._filter.bandpass import Bandpass
from starfish.image._filter.clip import Clip
from starfish.image._filter.gaussian_low_pass import GaussianLowPass
from starfish.spots._detector.trackpy_local_max_peak_finder import TrackpyLocalMaxPeakFinder


@pytest.mark.skip('issues with checksums prevent this data from working properly')
def test_allen_smFISH_cropped_data():

    # set random seed to errors provoked by optimization functions
    np.random.seed(777)

    # load the experiment
    experiment = starfish.data.allen_smFISH(use_test_data=True)

    primary_image = experiment.fov()[FieldOfView.PRIMARY_IMAGES]

    clip = Clip(p_min=10, p_max=100)
    clipped_image = clip.run(primary_image, in_place=False)

    bandpass = Bandpass(lshort=0.5, llong=7, threshold=None, truncate=4)
    bandpassed_image = bandpass.run(clipped_image, in_place=False)

    clip = Clip(p_min=10, p_max=100, is_volume=False)
    clipped_bandpassed_image = clip.run(bandpassed_image, in_place=False)

    sigma = (1, 0, 0)  # filter only in z, do nothing in x, y
    glp = GaussianLowPass(sigma=sigma, is_volume=True)
    z_filtered_image = glp.run(clipped_bandpassed_image, in_place=False)

    lmpf = TrackpyLocalMaxPeakFinder(
        spot_diameter=3,
        min_mass=300,
        max_size=3,
        separation=5,
        noise_size=0.65,
        preprocess=False,
        percentile=10,
        verbose=True,
        is_volume=True,
    )
    intensities = lmpf.run(z_filtered_image)  # noqa



from typing import Optional

import click

from starfish.imagestack.imagestack import ImageStack
from starfish.types import Indices
from ._base import FilterAlgorithmBase


class MaxProj(FilterAlgorithmBase):

    def __init__(self, dims) -> None:
        self.dims = dims

    _DEFAULT_TESTING_PARAMETERS = {"dims": 'r'}

    def run(
            self, stack: ImageStack, in_place: bool = False, verbose: bool = False,
            n_processes: Optional[int] = None
    ) -> ImageStack:
        return stack.max_proj(*tuple(Indices(dim) for dim in self.dims))

    @staticmethod
    @click.command("MaxProj")
    @click.option("--dims", type=str, multiple=True,
                  help="The dimensions the Imagestack should max project over. Options:"
                       "(r, c, z, y, or x) For multiple dimensions add multiple --dims. Ex."
                       "--dims r --dims c")
    @click.pass_context
    def _cli(ctx, dims):
        ctx.obj["component"]._cli_run(ctx, MaxProj(dims))



from typing import Optional

import click
import numpy as np
from skimage.morphology import ball, disk, white_tophat

from starfish.imagestack.imagestack import ImageStack
from ._base import FilterAlgorithmBase
from .util import determine_axes_to_group_by


class WhiteTophat(FilterAlgorithmBase):
    """
    Performs "white top hat" filtering of an image to enhance spots. "White top hat filtering"
    finds spots that are both smaller and brighter than their surroundings.

    See Also
    --------
    https://en.wikipedia.org/wiki/Top-hat_transform
    """

    def __init__(self, masking_radius: int, is_volume: bool=False) -> None:
        """
        Instance of a white top hat morphological masking filter which masks objects larger
        than `masking_radius`

        Parameters
        ----------
        masking_radius : int
            radius of the morphological masking structure in pixels
        is_volume : int
            If True, 3d (z, y, x) volumes will be filtered, otherwise, filter 2d tiles
            independently.

        """
        self.masking_radius = masking_radius
        self.is_volume = is_volume

    _DEFAULT_TESTING_PARAMETERS = {"masking_radius": 3}

    def _white_tophat(self, image: np.ndarray) -> np.ndarray:
        if self.is_volume:
            structuring_element = ball(self.masking_radius)
        else:
            structuring_element = disk(self.masking_radius)
        return white_tophat(image, selem=structuring_element)

    def run(
            self, stack: ImageStack, in_place: bool=False, verbose: bool=False,
            n_processes: Optional[int]=None
    ) -> ImageStack:
        """Perform filtering of an image stack

        Parameters
        ----------
        stack : ImageStack
            Stack to be filtered.
        in_place : bool
            if True, process ImageStack in-place, otherwise return a new stack
        verbose : bool
            If True, report on the percentage completed (default = False) during processing
        n_processes : Optional[int]
            Number of parallel processes to devote to calculating the filter

        Returns
        -------
        ImageStack :
            If in-place is False, return the results of filter as a new stack.  Otherwise return the
            original stack.

        """
        group_by = determine_axes_to_group_by(self.is_volume)
        result = stack.apply(
            self._white_tophat,
            group_by=group_by, verbose=verbose, in_place=in_place, n_processes=n_processes
        )
        return result

    @staticmethod
    @click.command("WhiteTophat")
    @click.option(
        "--masking-radius", default=15, type=int,
        help="diameter of morphological masking disk in pixels")
    @click.option(  # FIXME: was this intentionally missed?
        "--is-volume", is_flag=True, help="filter 3D volumes")
    @click.pass_context
    def _cli(ctx, masking_radius, is_volume):
        ctx.obj["component"]._cli_run(ctx, WhiteTophat(masking_radius, is_volume))



from typing import Tuple

import numpy as np

from starfish.codebook.codebook import Codebook
from starfish.imagestack.imagestack import ImageStack
from starfish.intensity_table.intensity_table import IntensityTable


class SyntheticData:
    """Synthetic generator for spot data and container objects

    Currently, this generator only generates codebooks with single channels "on" in each
    imaging round. Therefore, it can generate ISS and smFISH type experiments, but not
    more complex codes.

    Examples
    --------
    >>> from starfish.util.synthesize import SyntheticData

    >>> sd = SyntheticData(n_ch=3, n_round=4, n_codes=2)
    >>> codes = sd.codebook()
    >>> codes
    <xarray.Codebook (target: 2, c: 3, h: 4)>
    array([[[0, 0, 0, 0],
            [0, 0, 1, 1],
            [1, 1, 0, 0]],

           [[1, 0, 0, 0],
            [0, 0, 1, 0],
            [0, 1, 0, 1]]], dtype=uint8)
    Coordinates:
      * target     (target) object 08b1a822-a1b4-4e06-81ea-8a4bd2b004a9 ...
      * c          (c) int64 0 1 2
      * h          (h) int64 0 1 2 3

    >>> # note that the features are drawn from the codebook
    >>> intensities = sd.intensities(codebook=codes)
    >>> intensities
    <xarray.IntensityTable (features: 2, c: 3, h: 4)>
    array([[[    0.,     0.,     0.,     0.],
            [    0.,     0.,  8022., 12412.],
            [11160.,  9546.,     0.,     0.]],

           [[    0.,     0.,     0.,     0.],
            [    0.,     0., 10506., 10830.],
            [11172., 12331.,     0.,     0.]]])
    Coordinates:
      * features   (features) MultiIndex
      - z          (features) int64 7 3
      - y          (features) int64 14 32
      - x          (features) int64 32 15
      - r          (features) float64 nan nan
      * c          (c) int64 0 1 2
      * h          (h) int64 0 1 2 3
        target     (features) object 08b1a822-a1b4-4e06-81ea-8a4bd2b004a9 ...

    >>> sd.spots(intensities=intensities)
    <starfish.image._stack.ImageStack at 0x10a60c5f8>

    """

    def __init__(
            self,
            n_round: int=4,
            n_ch: int=4,
            n_z: int=10,
            height: int=50,
            width: int=45,
            n_codes: int=16,
            n_spots: int=20,
            n_photons_background: int=1000,
            background_electrons: int=1,
            point_spread_function: Tuple[int, ...]=(4, 2, 2),
            camera_detection_efficiency: float=0.25,
            gray_level: float=37000.0 / 2 ** 16,
            ad_conversion_bits: int=16,
            mean_fluor_per_spot: int=200,
            mean_photons_per_fluor: int=50,

    ) -> None:
        self.n_round = n_round
        self.n_ch = n_ch
        self.n_z = n_z
        self.height = height
        self.width = width
        self.n_codes = n_codes
        self.n_spots = n_spots
        self.n_photons_background = n_photons_background
        self.background_electrons = background_electrons
        self.point_spread_function = point_spread_function
        self.camera_detection_efficiency = camera_detection_efficiency
        self.gray_level = gray_level
        self.ad_coversion_bits = ad_conversion_bits
        self.mean_fluor_per_spot = mean_fluor_per_spot
        self.mean_photons_per_fluor = mean_photons_per_fluor

    def codebook(self) -> Codebook:
        return Codebook.synthetic_one_hot_codebook(self.n_round, self.n_ch, self.n_codes)

    def intensities(self, codebook=None) -> IntensityTable:
        if codebook is None:
            codebook = self.codebook()
        intensities = IntensityTable.synthetic_intensities(
            codebook, self.n_z, self.height, self.width, self.n_spots,
            self.mean_fluor_per_spot, self.mean_photons_per_fluor)
        assert intensities.dtype == np.float32 and intensities.max() <= 1
        return intensities

    def spots(self, intensities=None) -> ImageStack:
        if intensities is None:
            intensities = self.intensities()
        stack = ImageStack.synthetic_spots(
            intensities, self.n_z, self.height, self.width, self.n_photons_background,
            self.point_spread_function, self.camera_detection_efficiency,
            self.background_electrons, self.gray_level, self.ad_coversion_bits)
        assert stack.xarray.dtype == np.float32
        return stack



import click

from starfish.codebook.codebook import Codebook
from starfish.intensity_table.intensity_table import IntensityTable
from ._base import DecoderAlgorithmBase


class PerRoundMaxChannelDecoder(DecoderAlgorithmBase):

    def __init__(self):
        pass

    def run(self, intensities: IntensityTable, codebook: Codebook) -> IntensityTable:
        """Decode spots by selecting the max-valued channel in each sequencing round

        Parameters
        ----------
        intensities : IntensityTable
            IntensityTable to be decoded
        codebook : Codebook
            Contains codes to decode IntensityTable

        Returns
        -------
        IntensityTable :
            IntensityTable decoded and appended with Features.TARGET and Features.QUALITY values.

        """
        return codebook.decode_per_round_max(intensities)

    @staticmethod
    @click.command("PerRoundMaxChannelDecoder")
    @click.pass_context
    def _cli(ctx):
        ctx.obj["component"]._cli_run(ctx, PerRoundMaxChannelDecoder())



from typing import Any, Dict, List, Type

import click
from skimage.io import imsave

from starfish.imagestack.imagestack import ImageStack
from starfish.pipeline import AlgorithmBase, PipelineComponent
from . import watershed
from ._base import SegmentationAlgorithmBase


class Segmentation(PipelineComponent):

    @classmethod
    def _get_algorithm_base_class(cls) -> Type[AlgorithmBase]:
        return SegmentationAlgorithmBase

    @classmethod
    def _cli_run(cls, ctx, instance):
        output = ctx.obj["output"]
        pri_stack = ctx.obj["primary_images"]
        nuc_stack = ctx.obj["nuclei"]

        label_image = instance.run(pri_stack, nuc_stack)

        print(f"Writing label image to {output}")
        imsave(output, label_image)

    @staticmethod
    @click.group("segment")
    @click.option("--primary-images", required=True, type=click.Path(exists=True))
    @click.option("--nuclei", required=True, type=click.Path(exists=True))
    @click.option("-o", "--output", required=True)
    @click.pass_context
    def _cli(ctx, primary_images, nuclei, output):
        print('Segmenting ...')
        ctx.obj = dict(
            component=Segmentation,
            output=output,
            primary_images=ImageStack.from_path_or_url(primary_images),
            nuclei=ImageStack.from_path_or_url(nuclei),
        )


Segmentation._cli_register()



import json
import os
from typing import Any, Dict, Sequence, Union


class Config(object):

    __NO_VALUE_PASSED = object()

    def __init__(self, value: Union[str, Dict]=None,
                 key: str="STARFISH_CONFIG") -> None:
        """
        Parse user arguments, environment variables, and external files to
        generate a configuration object.

        Parameters
        ----------
        value: Union[str, Dict]
            Either a json-object-like structure which will be passed unmodified to the
            Config constructor, or a string which will be used to build such a structure. If
            the string starts with an "@", then it will be interpreted as the filename of a
            json file which should be read. Otherwise, it will be parsed as a json string.
        key: str
            The name of an environment variable which will be used
            if the value is None. STARFISH_CONFIG will be used if no
            key is provided.
        """
        # Record the original values
        self.__value = value
        self.__key = key

        if value is None:
            value = os.environ.get(key)

        if not value:
            value = {}

        if isinstance(value, str):
            if value.startswith("@"):
                with open(value[1:], "r") as o:
                    data = json.loads(o.read())
            else:
                data = json.loads(value)
        else:
            data = value

        self.data = data

    def lookup(self, keys: Sequence[str], value: Any=__NO_VALUE_PASSED) -> Any:
        """
        Parameters
        ----------
        keys: Sequence[str]
            Sequence of keys which will be looked up one after the other.
            If any key does not exist or None is returned, then the provided
            default value will be returned. If no default value is provided,
            an exception is raised.
        value: Any
            Default value to return in the case that the lookup fails. An
            exception will be raised if none is provided.
        """
        data: Any = self.data
        for key in keys:
            try:
                data = data.get(key)
            except (AttributeError, KeyError):
                data = None  # Clear value
                if value is Config.__NO_VALUE_PASSED:
                    raise
        if data:
            # If we've reached here without exception,
            # then we've found the value.
            return data
        elif value is not Config.__NO_VALUE_PASSED:
            return value
        else:
            raise KeyError(keys)



"""
This module stores a contract for new Filter algorithms.

Contract
--------

- constructor accepts 3d and 2d data, where default is 2d
- values emitted by a filter are floats between 0 and 1 (inclusive)
- exposes a `run`() method
- run accepts an in-place parameter which defaults to True
- run always returns an ImageStack (if in-place, returns a reference to the modified input data)
- run accepts an `n_processes` parameter which determines
- run accepts a `verbose` parameter, which triggers tqdm to print progress

To add a new filter, simply add default
parameters for the constructor (omit is_volume), and it will be tested against the contract by
way of registration to the FilterAlgorithmBase
"""

from typing import Dict

import numpy as np
import pytest

from starfish import ImageStack
from starfish.image import Filter
from starfish.image._filter.max_proj import MaxProj

methods: Dict = Filter._algorithm_to_class_map()


def generate_default_data():
    data = np.random.rand(2, 2, 2, 40, 50).astype(np.float32)
    return ImageStack.from_numpy_array(data)


@pytest.mark.parametrize('filter_class', methods.values())
def test_all_methods_adhere_to_contract(filter_class):
    """Test that all filter algorithms adhere to the filtering contract"""

    default_kwargs = filter_class._DEFAULT_TESTING_PARAMETERS

    # TODO ambrosejcarr: all filters should accept boolean is_volume
    instance = filter_class(**default_kwargs)

    # TODO ambrosejcarr: all methods should process 3d data
    # # stores is_volume, and is_volume is a bool
    # try:
    #     volume_param = getattr(instance, 'is_volume')
    # except AttributeError:
    #     raise AttributeError(f'{filter_class} should accept and store is_volume.')
    #
    # assert isinstance(volume_param, bool), \
    #     f'{filter_class} is_volume must be a bool, not {type(volume_param)}'

    # always emits an Image, even if in_place=True and the resulting filter operates in-place
    data = generate_default_data()
    try:
        filtered = instance.run(data, in_place=True)
    except TypeError:
        raise AssertionError(f'{filter_class} must accept in_place parameter')
    assert isinstance(filtered, ImageStack)
    if filter_class is not MaxProj:
        # Max Proj does not have an in place option, so we need to skip this assertion
        assert data is filtered, \
            f'{filter_class} should return a reference to the input ImageStack when run in_place'

    # operates out of place
    data = generate_default_data()
    filtered = instance.run(data, in_place=False)
    assert data is not filtered, \
        f'{filter_class} should output a new ImageStack when run out-of-place'

    # accepts n_processes
    # TODO shanaxel: verify that this causes more than one process to be generated
    data = generate_default_data()
    try:
        instance.run(data, n_processes=1)
    except TypeError:
        raise AssertionError(f'{filter_class} must accept n_processes parameter')

    # accepts verbose, and if passed, prints progress
    data = generate_default_data()
    try:
        instance.run(data, verbose=True)
    except TypeError:
        raise AssertionError(f'{filter_class} must accept verbose parameter')

    # output is dtype float and within the expected interval of [0, 1]
    assert filtered.xarray.dtype == np.float32, f'{filter_class} must output float32 data'
    assert np.all(filtered.xarray >= 0), \
        f'{filter_class} must output a result where all values are >= 0'
    assert np.all(filtered.xarray <= 1), \
        f'{filter_class} must output a result where all values are <= 1'



from typing import Tuple

import click
import numpy as np

from starfish.codebook.codebook import Codebook
from starfish.imagestack.imagestack import ImageStack
from starfish.intensity_table.intensity_table import IntensityTable
from ._base import SpotFinderAlgorithmBase
from .combine_adjacent_features import CombineAdjacentFeatures, ConnectedComponentDecodingResult


class PixelSpotDetector(SpotFinderAlgorithmBase):
    def __init__(
            self, codebook: Codebook, metric: str, distance_threshold: float,
            magnitude_threshold: int, min_area: int, max_area: int, norm_order: int=2,
            crop_x: int=0, crop_y: int=0, crop_z: int=0) -> None:
        """Decode an image by first coding each pixel, then combining the results into spots

        Parameters
        ----------
        codebook : Codebook
            Codebook object mapping codewords to the targets they are designed to detect
        metric : str
            the sklearn metric string to pass to NearestNeighbors
        distance_threshold : float
            spots whose codewords are more than this metric distance from an expected code are
            filtered
        magnitude_threshold : int
            spots with intensity less than this value are filtered
        min_area : int
            spots with total area less than this value are filtered
        max_area : int
            spots with total area greater than this value are filtered
        norm_order : int
            order of L_p norm to apply to intensities and codes when using metric_decode to pair
            each intensities to its closest target (default = 2)
        crop_x, crop_y, crop_z : int
            number of pixels to crop from the top and bottom of each of the x, y, and z axes of
            an ImageStack (default = 0)

        """
        self.codebook = codebook
        self.metric = metric
        self.distance_threshold = distance_threshold
        self.magnitude_threshold = magnitude_threshold
        self.min_area = min_area
        self.max_area = max_area
        self.norm_order = norm_order
        self.crop_x = crop_x
        self.crop_y = crop_y
        self.crop_z = crop_z

    def run(
        self, stack: ImageStack,
    ) -> Tuple[IntensityTable, ConnectedComponentDecodingResult]:
        """decode pixels and combine them into spots using connected component labeling

        Parameters
        ----------
        stack : ImageStack
            ImageStack containing spots

        Returns
        -------
        IntensityTable :
            IntensityTable containing decoded spots
        ConnectedComponentDecodingResult :
            Results of connected component labeling

        """
        pixel_intensities = IntensityTable.from_image_stack(
            stack, crop_x=self.crop_x, crop_y=self.crop_y, crop_z=self.crop_z)
        decoded_intensities = self.codebook.metric_decode(
            pixel_intensities,
            max_distance=self.distance_threshold,
            min_intensity=self.magnitude_threshold,
            norm_order=self.norm_order,
            metric=self.metric
        )
        caf = CombineAdjacentFeatures(
            min_area=self.min_area,
            max_area=self.max_area,
            mask_filtered_features=True
        )
        decoded_spots, image_decoding_results = caf.run(intensities=decoded_intensities)

        return decoded_spots, image_decoding_results

    @staticmethod
    @click.command("PixelSpotDetector")
    @click.option("--metric", type=str, default='euclidean')
    @click.option(
        "--distance-threshold", type=float, default=0.5176,
        help="maximum distance a pixel may be from a codeword before it is filtered"
    )
    @click.option(
        "--magnitude-threshold", type=float, default=1,
        help="minimum magnitude of a feature"
    )
    @click.option(
        "--min-area", type=int, default=2,
        help="minimum area of a feature"
    )
    @click.option(
        "--max-area", type=int, default=np.inf,
        help="maximum area of a feature"
    )
    @click.option(
        "--norm-order", type=int, default=2,
        help="order of L_p norm to apply to intensities "
        "and codes when using metric_decode to pair each intensities to its closest target"
    )
    @click.option('--crop-x', type=int, default=0)
    @click.option('--crop-y', type=int, default=0)
    @click.option('--crop-z', type=int, default=0)
    @click.pass_context
    def _cli(ctx, metric, distance_threshold, magnitude_threshold,
             min_area, max_area, norm_order, crop_x, crop_y, crop_z):
        codebook = ctx.obj["codebook"]
        instance = PixelSpotDetector(
            codebook=codebook,
            metric=metric,
            distance_threshold=distance_threshold,
            magnitude_threshold=magnitude_threshold,
            min_area=min_area,
            max_area=max_area,
            norm_order=norm_order,
            crop_x=crop_x,
            crop_y=crop_y,
            crop_z=crop_z
        )
        ctx.obj["component"]._cli_run(ctx, instance)



import os
import sys
import unittest

from starfish.test.full_pipelines.cli._base_cli_test import CLITest

@unittest.skip("skipping for now")
class TestWithMerfishData(CLITest, unittest.TestCase):
    # __test__ = True
    # TODO this test currently fails because it doesn't do the scaling the notebook does in memory
    # Need to figure out ScaleByPercentile change

    @property
    def subdirs(self):
        return (
            "registered",
            "filtered",
            "results"
        )

    @property
    def stages(self):
        return (
            [
                sys.executable,
                "examples/get_cli_test_data.py",
                "https://d2nhj9g34unfro.cloudfront.net/20181005/MERFISH-TEST/experiment.json",
                lambda tempdir, *args, **kwargs: os.path.join(tempdir, "registered")
            ],
            [
                "starfish", "filter",
                "--input", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "registered/fov_001", "hybridization.json"),
                "--output", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "filtered", "gaussian_filtered.json"),
                "GaussianHighPass",
                "--sigma", "3",
            ],
            [
                "starfish", "filter",
                "--input", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "filtered", "gaussian_filtered.json"),
                "--output", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "filtered", "deconvolve_filtered.json"),
                "DeconvolvePSF",
                "--sigma", "2",
                "--num-iter", "9"
            ],
            [
                "starfish", "filter",
                "--input", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "filtered", "deconvolve_filtered.json"),
                "--output", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "filtered", "gaussian_filtered.json"),
                "GaussianLowPass",
                "--sigma", "1"
            ],
            [
                "starfish", "filter",
                "--input", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "filtered", "deconvolve_filtered.json"),
                "--output", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "filtered", "scale_filtered.json"),
                "ScaleByPercentile",
                "--p", "90",
            ],
            [
                "starfish", "detect_spots",
                "--input", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "filtered", "scale_filtered.json"),
                "--output", lambda tempdir, *args, **kwargs: os.path.join(tempdir, "results"),
                "PixelSpotDetector",
                "--codebook", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "registered", "codebook.json"),
                "--distance-threshold", "0.5176",
                "--magnitude-threshold", "5e-5",
                "--norm-order", "2",
                "--crop-x", "0",
                "--crop-y", "40",
                "--crop-z", "40"
            ],
        )

    def verify_results(self, intensities):
        pass



import pytest
from pkg_resources import resource_filename

from .util import SpaceTxValidator

package_name = "sptx_format"
fov_schema_path = resource_filename(package_name, "schema/field_of_view/field_of_view.json")
validator = SpaceTxValidator(fov_schema_path)
example = resource_filename(package_name, "examples/field_of_view/field_of_view.json")


def test_field_of_view():
    assert validator.validate_file(example)


def test_dartfish_example_field_of_view():
    dartfish_example_path = resource_filename(
        package_name, "examples/field_of_view/dartfish_field_of_view.json")
    assert validator.validate_file(dartfish_example_path)


def test_dartfish_nuclei_example_field_of_view():
    dartfish_example_path = resource_filename(
        package_name, "examples/field_of_view/dartfish_nuclei.json")
    assert validator.validate_file(dartfish_example_path)


def test_channel_must_be_present():
    no_channel = validator.load_json(example)
    del no_channel['tiles'][0]['indices']['c']
    with pytest.warns(UserWarning):
        assert not validator.validate_object(no_channel)


def test_round_must_be_present():
    mangled_round = validator.load_json(example)
    del mangled_round['tiles'][0]['indices']['r']
    mangled_round['tiles'][0]['indices']['h'] = 0
    with pytest.warns(UserWarning):
        assert not validator.validate_object(mangled_round)



from typing import Callable, Sequence, Tuple, Union

import numpy as np
import xarray as xr

from starfish.imagestack.imagestack import ImageStack
from starfish.intensity_table.intensity_table import IntensityTable
from starfish.pipeline.algorithmbase import AlgorithmBase
from starfish.types import Number, SpotAttributes
from .combine_adjacent_features import ConnectedComponentDecodingResult


class SpotFinderAlgorithmBase(AlgorithmBase):
    def run(
            self,
            hybridization_image: ImageStack,
    ) -> Union[IntensityTable, Tuple[IntensityTable, ConnectedComponentDecodingResult]]:
        """Finds spots in an ImageStack"""
        raise NotImplementedError()

    def image_to_spots(self, data_image: Union[np.ndarray, xr.DataArray]) -> SpotAttributes:
        """Finds spots in a 3d volume"""
        raise NotImplementedError()

    @staticmethod
    def _get_measurement_function(measurement_type: str) -> Callable[[Sequence], Number]:
        try:
            measurement_function = getattr(np, measurement_type)
        except AttributeError:
            raise ValueError(
                f'measurement_type must be a numpy reduce function such as "max" or "mean". '
                f'{measurement_type} not found.')
        return measurement_function



from enum import Enum


class AugmentedEnum(Enum):
    def __hash__(self):
        return self.value.__hash__()

    def __eq__(self, other):
        if isinstance(other, type(self)) or isinstance(other, str):
            return self.value == other
        return False

    def __str__(self) -> str:
        return self.value


class Coordinates(AugmentedEnum):
    Z = 'zc'
    Y = 'yc'
    X = 'xc'


PHYSICAL_COORDINATE_DIMENSION = "physical_coordinate"
"""
This is the xarray dimension name for the physical coordinates of the tiles.
"""


class PhysicalCoordinateTypes(AugmentedEnum):
    """
    These are more accurately the xarray coordinates for the physical coordinates of a tile.
    """
    Z_MAX = 'zmax'
    Z_MIN = 'zmin'
    Y_MAX = 'ymax'
    Y_MIN = 'ymin'
    X_MAX = 'xmax'
    X_MIN = 'xmin'


class Indices(AugmentedEnum):
    ROUND = 'r'
    CH = 'c'
    Z = 'z'
    Y = 'y'
    X = 'x'


class Features:
    """
    contains constants relating to the codebook and feature (spot/pixel) representations of the
    image data
    """

    AXIS = 'features'
    TARGET = 'target'
    CODEWORD = 'codeword'
    CODE_VALUE = 'v'
    SPOT_RADIUS = 'radius'
    DISTANCE = 'distance'
    PASSES_THRESHOLDS = 'passes_thresholds'
    CELL_ID = 'cell_id'
    SPOT_ID = 'spot_id'
    INTENSITY = 'intensity'



"""
This module contains default style constants for visualization methods in starfish
"""

import matplotlib.pyplot as plt

TARGETS_COLORMAP = plt.get_cmap('nipy_spectral')

BACKGROUND_COLORMAP = plt.get_cmap('gray')

# TODO ambrosejcarr: propagate this through plots
FONTSIZE = 14



from itertools import product
from typing import Dict, Optional, Union

import numpy as np
import pandas as pd
import regional
import xarray as xr

from starfish.expression_matrix.expression_matrix import ExpressionMatrix
from starfish.image._filter.util import preserve_float_range
from starfish.types import Features, Indices, SpotAttributes


class IntensityTable(xr.DataArray):
    """Container for spot/pixel features extracted from image data

    An IntensityTable is comprised of each feature's intensity across channels and imaging
    rounds, where features are typically spots or pixels. This forms an
    ``(n_feature, n_channel, n_round)`` tensor implemented as an xarray.DataArray object.
    In addition to the basic xarray methods, IntensityTable implements:

    Methods
    -------
    empty_intensity_table(spot_attributes, n_ch, n_round)
        creates an IntensityTable with all intensities equal to zero

    from_spot_data(intensities, spot_attributes)
        creates an IntensityTable from a 3d array and a spot attributes dataframe

    synthetic_intensities(codebook, num_z=12, height=50, width=40, n_spots=10, \
            mean_fluor_per_spot=200, mean_photons_per_fluor=50)
        creates an IntensityTable with synthetic spots, given a codebook

    save(filename)
        save the IntensityTable to netCDF

    load(filename)
        load an IntensityTable from netCDF

    Examples
    --------
    Create an IntensityTable using the ``synthetic_intensities`` method::

        >>> from starfish.util.synthesize import SyntheticData
        >>> sd = SyntheticData(n_ch=3, n_round=4, n_codes=2)
        >>> codes = sd.codebook()
        >>> sd.intensities(codebook=codes)
        <xarray.IntensityTable (features: 2, c: 3, h: 4)>
        array([[[    0.,     0.,     0.,     0.],
                [    0.,     0.,  8022., 12412.],
                [11160.,  9546.,     0.,     0.]],

               [[    0.,     0.,     0.,     0.],
                [    0.,     0., 10506., 10830.],
                [11172., 12331.,     0.,     0.]]])
        Coordinates:
        * features   (features) MultiIndex
        - z          (features) int64 7 3
        - y          (features) int64 14 32
        - x          (features) int64 32 15
        - r          (features) float64 nan nan
        * c          (c) int64 0 1 2
        * h          (h) int64 0 1 2 3
          target     (features) object 08b1a822-a1b4-4e06-81ea-8a4bd2b004a9 ...

    """

    @staticmethod
    def _build_xarray_coords(
            spot_attributes: SpotAttributes, channel_index: np.ndarray, round_index: np.ndarray
    ) -> Dict[str, np.ndarray]:
        """build a non-multi-index set of coordinates for intensity-table"""
        coordinates = {
            k: (Features.AXIS, spot_attributes.data[k].values)
            for k in spot_attributes.data}
        coordinates.update({
            Features.AXIS: np.arange(len(spot_attributes.data)),
            Indices.CH.value: channel_index,
            Indices.ROUND.value: round_index
        })
        return coordinates

    @classmethod
    def empty_intensity_table(
            cls, spot_attributes: SpotAttributes, n_ch: int, n_round: int,
    ) -> "IntensityTable":
        """Create an empty intensity table with pre-set axis whose values are zero

        Parameters
        ----------
        spot_attributes : pd.MultiIndex
            MultiIndex containing spot metadata. Must contain the values specifid in Constants.X,
            Y, Z, and RADIUS.
        n_ch : int
            number of channels measured in the imaging experiment
        n_round : int
            number of imaging rounds measured in the imaging experiment

        Returns
        -------
        IntensityTable :
            empty IntensityTable

        """
        if not isinstance(spot_attributes, SpotAttributes):
            raise TypeError('parameter spot_attributes must be a starfish SpotAttributes object.')

        channel_index = np.arange(n_ch)
        round_index = np.arange(n_round)
        data = np.zeros((spot_attributes.data.shape[0], n_ch, n_round))
        dims = (Features.AXIS, Indices.CH.value, Indices.ROUND.value)
        coords = cls._build_xarray_coords(spot_attributes, channel_index, round_index)

        intensity_table = cls(
            data=data, coords=coords, dims=dims,
        )

        return intensity_table

    @classmethod
    def from_spot_data(
            cls, intensities: Union[xr.DataArray, np.ndarray], spot_attributes: SpotAttributes,
            *args, **kwargs) -> "IntensityTable":
        """Table to store image feature intensities and associated metadata

        Parameters
        ----------
        intensities : np.ndarray[Any]
            intensity data
        spot_attributes : pd.DataFrame
            Name(s) of the data dimension(s). Must be either a string (only
            for 1D data) or a sequence of strings with length equal to the
            number of dimensions. If this argument is omitted, dimension names
            are taken from ``coords`` (if possible) and otherwise default to
            ``['dim_0', ... 'dim_n']``.
        args :
            additional arguments to pass to the xarray constructor
        kwargs :
            additional keyword arguments to pass to the xarray constructor

        Returns
        -------
        IntensityTable :
            IntensityTable containing data from passed ndarray, annotated by spot_attributes

        """

        if len(intensities.shape) != 3:
            raise ValueError(
                f'intensities must be a (features * ch * round) 3-d tensor. Provided intensities '
                f'shape ({intensities.shape}) is invalid.')

        if not isinstance(spot_attributes, SpotAttributes):
            raise TypeError('parameter spot_attributes must be a starfish SpotAttributes object.')

        coords = cls._build_xarray_coords(
            spot_attributes,
            np.arange(intensities.shape[1]),
            np.arange(intensities.shape[2]))

        dims = (Features.AXIS, Indices.CH.value, Indices.ROUND.value)

        intensities = cls(intensities, coords, dims, *args, **kwargs)
        return intensities

    def save(self, filename: str) -> None:
        """Save an IntensityTable as a Netcdf File

        Parameters
        ----------
        filename : str
            Name of Netcdf file

        """
        self.to_netcdf(filename)

    def save_mermaid(self, filename: str) -> pd.DataFrame:
        """
        Writes a .csv.gz file in columnar format that is readable by MERMAID visualization
        software.

        To run MERMAID, follow the installation instructions for that repository and simply
        replace the data.csv.gz file with the output of this function.

        Parameters
        ----------
        filename : str
            name for compressed-gzipped MERMAID data file. Should end in '.csv.gz'

        See Also
        --------
        https://github.com/JEFworks/MERmaid

        """

        # verify the IntensityTable has been decoded
        if Features.TARGET not in self.coords.keys():
            raise RuntimeError(
                'IntensityTable must be decoded before it can be converted to MERMAID input.'
            )

        # construct the MERMAID dataframe. As MERMAID adds support for non-categorical variables,
        # additional columns can be added here
        df = self.to_features_dataframe()
        column_order = [
            Indices.X,
            Indices.Y,
            Features.TARGET,
            Features.TARGET,  # added twice to support simultaneous coding
        ]
        mermaid_data = df[column_order]

        # write to disk
        mermaid_data.to_csv(filename, compression='gzip', index=False)

    @classmethod
    def load(cls, filename: str) -> "IntensityTable":
        """load an IntensityTable from Netcdf

        Parameters
        ----------
        filename : str
            File to load

        Returns
        -------
        IntensityTable

        """
        loaded = xr.open_dataarray(filename)
        intensity_table = cls(
            loaded.data,
            loaded.coords,
            loaded.dims
        )
        return intensity_table

    def show(self, background_image: np.ndarray) -> None:
        """show spots on a background image"""
        raise NotImplementedError

    @classmethod
    def synthetic_intensities(
            cls, codebook, num_z: int=12, height: int=50, width: int=40, n_spots=10,
            mean_fluor_per_spot=200, mean_photons_per_fluor=50
    ) -> "IntensityTable":
        """Create an IntensityTable containing synthetic spots with random locations

        Parameters
        ----------
        codebook : Codebook
            starfish codebook object
        num_z :
            number of z-planes to use when localizing spots
        height :
            y dimension of each synthetic plane
        width :
            x dimension of each synthetic plane
        n_spots :
            number of spots to generate
        mean_fluor_per_spot :
            mean number of fluorophores per spot
        mean_photons_per_fluor :
            mean number of photons per fluorophore.

        Returns
        -------
        IntensityTable

        """

        # TODO nsofroniew: right now there is no jitter on x-y positions of the spots
        z = np.random.randint(0, num_z, size=n_spots)
        y = np.random.randint(0, height, size=n_spots)
        x = np.random.randint(0, width, size=n_spots)
        r = np.empty(n_spots)
        r.fill(np.nan)  # radius is a function of the point-spread gaussian size
        spot_attributes = SpotAttributes(
            pd.DataFrame(
                {Indices.Z.value: z,
                 Indices.Y.value: y,
                 Indices.X.value: x,
                 Features.SPOT_RADIUS: r}
            )
        )

        # empty data tensor
        data = np.zeros(shape=(n_spots, *codebook.shape[1:]))

        targets = np.random.choice(
            codebook.coords[Features.TARGET], size=n_spots, replace=True)
        expected_bright_locations = np.where(codebook.loc[targets])

        # create a binary matrix where "on" spots are 1
        data[expected_bright_locations] = 1

        # add physical properties of fluorescence
        data *= np.random.poisson(mean_photons_per_fluor, size=data.shape)
        data *= np.random.poisson(mean_fluor_per_spot, size=data.shape)

        # convert data to float for consistency with starfish
        data = preserve_float_range(data)
        assert 0 < data.max() <= 1

        intensities = cls.from_spot_data(data, spot_attributes)
        intensities[Features.TARGET] = (Features.AXIS, targets)

        return intensities

    @classmethod
    def from_image_stack(
            cls,
            image_stack,
            crop_x: int=0, crop_y: int=0, crop_z: int=0
    ) -> "IntensityTable":
        """Generate an IntensityTable from all the pixels in the ImageStack

        Parameters
        ----------
        crop_x : int
            number of pixels to crop from both top and bottom of x
        crop_y : int
            number of pixels to crop from both top and bottom of y
        crop_z : int
            number of pixels to crop from both top and bottom of z
        image_stack : ImageStack
            ImageStack containing pixels to be treated as intensities

        Returns
        -------
        IntensityTable :
            IntensityTable containing one intensity per pixel (across channels and rounds)

        """

        # verify the image is large enough to crop
        assert crop_z * 2 < image_stack.shape['z']
        assert crop_y * 2 < image_stack.shape['y']
        assert crop_x * 2 < image_stack.shape['x']

        zmin = crop_z
        ymin = crop_y
        xmin = crop_x
        zmax = image_stack.shape['z'] - crop_z
        ymax = image_stack.shape['y'] - crop_y
        xmax = image_stack.shape['x'] - crop_x
        cropped_stack = image_stack.sel({Indices.Z: (zmin, zmax),
                                         Indices.Y: (ymin, ymax),
                                         Indices.X: (xmin, xmax)})

        data = cropped_stack.xarray.transpose(
            Indices.Z.value,
            Indices.Y.value,
            Indices.X.value,
            Indices.CH.value,
            Indices.ROUND.value,
        )

        # (pixels, ch, round)
        intensity_data = data.values.reshape(
            -1, image_stack.num_chs, image_stack.num_rounds)

        # IntensityTable pixel coordinates
        z = np.arange(zmin, zmax)
        y = np.arange(ymin, ymax)
        x = np.arange(xmin, xmax)

        feature_attribute_data = pd.DataFrame(
            data=np.array(list(product(z, y, x))),
            columns=['z', 'y', 'x']
        )
        feature_attribute_data[Features.SPOT_RADIUS] = np.full(
            feature_attribute_data.shape[0], fill_value=0.5
        )

        pixel_coordinates = SpotAttributes(feature_attribute_data)

        return IntensityTable.from_spot_data(intensity_data, pixel_coordinates)

    def to_features_dataframe(self) -> pd.DataFrame:
        """Generates a dataframe of the underlying features multi-index.
        This is guaranteed to contain the features x, y, z, and radius.

        Returns
        -------
        pd.DataFrame

        """
        return pd.DataFrame(dict(self[Features.AXIS].coords))

    def to_expression_matrix(self, regions: Optional[regional.many]=None) -> ExpressionMatrix:
        """Generates a cell x gene count matrix where each cell is annotated with spatial metadata

        Parameters
        ----------
        regions: Optional[regional.Many]
            cell segmentation results that were used to assign points to cells. If not provided, the
            centers of the cells will be estimated by taking the midpoint between the extreme-valued
            spots on each axis.

        Returns
        -------
        ExpressionMatrix :
            cell x gene expression table
        """

        # create the 2-d counts matrix
        grouped = self.to_features_dataframe().groupby(['cell_id', 'target'])
        counts = grouped.count().iloc[:, 0].unstack().fillna(0)

        if regions:
            # counts.index stores cell_id, extract cell information from the regional.many object
            metadata = {
                "area": ("cells", [regions[id_].area for id_ in counts.index]),
                "x": ("cells", [regions[id_].center[0] for id_ in counts.index]),
                "y": ("cells", [regions[id_].center[1] for id_ in counts.index]),
                "z": ("cells", np.zeros(counts.shape[0]))
            }
        else:
            grouped = self.to_features_dataframe().groupby(['cell_id'])[['x', 'y', 'z']]
            min_ = grouped.min()
            max_ = grouped.max()
            coordinate_df = min_ + (max_ - min_) / 2
            metadata = {name: ("cells", data.values) for name, data in coordinate_df.items()}
            metadata['area'] = ("cells", np.full(counts.shape[0], fill_value=np.nan))

        # add genes to the metadata
        metadata.update({"genes": counts.columns.values})
        metadata.update({"cell_id": ("cells", counts.index.values)})

        mat = ExpressionMatrix(
            data=counts.values,
            dims=('cells', 'genes'),
            coords=metadata,
            name='expression_matrix'
        )
        return mat

    def feature_trace_magnitudes(self) -> np.ndarray:
        """Return the magnitudes of each feature across rounds and channels

        Returns
        -------
        np.ndarray :
            vector of feature norms

        """
        feature_traces = self.stack(traces=(Indices.CH.value, Indices.ROUND.value))
        norm = np.linalg.norm(feature_traces.values, ord=2, axis=1)  # 2 = feature magnitudes

        return norm



from typing import Type

import click

from starfish.imagestack.imagestack import ImageStack
from starfish.pipeline import AlgorithmBase, PipelineComponent
from . import _base
from . import bandpass
from . import clip
from . import gaussian_high_pass
from . import gaussian_low_pass
from . import laplace
from . import max_proj
from . import mean_high_pass
from . import richardson_lucy_deconvolution
from . import scale_by_percentile
from . import white_tophat
from . import zero_by_channel_magnitude


class Filter(PipelineComponent):

    @classmethod
    def _get_algorithm_base_class(cls) -> Type[AlgorithmBase]:
        return _base.FilterAlgorithmBase

    @classmethod
    def _cli_run(cls, ctx, instance):
        output = ctx.obj["output"]
        stack = ctx.obj["stack"]
        filtered = instance.run(stack)
        filtered.write(output)

    @staticmethod
    @click.group("filter")
    @click.option("-i", "--input", type=click.Path(exists=True))
    @click.option("-o", "--output", required=True)
    @click.pass_context
    def _cli(ctx, input, output):
        print("Filtering images...")
        ctx.obj = dict(
            component=Filter,
            input=input,
            output=output,
            stack=ImageStack.from_path_or_url(input),
        )


Filter._cli_register()



import os
import sys

import starfish


ROOT_DIR = os.path.dirname(os.path.dirname(os.path.abspath(starfish.__file__)))
os.environ["TESTING"] = "1"
sys.path.append(os.path.join(ROOT_DIR, "docs", "source", "usage", "iss"))


def test_iss_pipeline_in_docs():
    # Just importing the file and verifying it runs for now
    __import__('iss_pipeline')



import os
from typing import Type

import click

from starfish.codebook.codebook import Codebook
from starfish.imagestack.imagestack import ImageStack
from starfish.pipeline import AlgorithmBase, PipelineComponent
from starfish.types import Indices
from . import _base
from . import blob
from . import pixel_spot_detector
from . import trackpy_local_max_peak_finder


class SpotFinder(PipelineComponent):

    @classmethod
    def _get_algorithm_base_class(cls) -> Type[AlgorithmBase]:
        return _base.SpotFinderAlgorithmBase

    @classmethod
    def _cli_run(cls, ctx, instance):
        output = ctx.obj["output"]
        blobs_stack = ctx.obj["blobs_stack"]
        image_stack = ctx.obj["image_stack"]
        ref_image = ctx.obj["reference_image_from_max_projection"]
        if blobs_stack is not None:
            blobs_stack = ImageStack.from_path_or_url(blobs_stack)  # type: ignore
            mp = blobs_stack.max_proj(Indices.ROUND, Indices.CH)
            mp_numpy = mp._squeezed_numpy(Indices.ROUND, Indices.CH)
            #  TODO: this won't work for PixelSpotDectector
            intensities = instance.run(
                image_stack,
                blobs_image=mp_numpy,
                reference_image_from_max_projection=ref_image,
            )
        else:
            intensities = instance.run(image_stack)

        # When PixelSpotDetector is used run() returns a tuple
        if isinstance(intensities, tuple):
            intensities = intensities[0]
        intensities.save(output)

    @staticmethod
    @click.group("detect_spots")
    @click.option("-i", "--input", required=True, type=click.Path(exists=True))
    @click.option("-o", "--output", required=True)
    @click.option(
        '--blobs-stack', default=None, required=False, help=(
            'ImageStack that contains the blobs. Will be max-projected across imaging round '
            'and channel to produce the blobs_image'
        )
    )
    @click.option(
        '--reference-image-from-max-projection', default=False, is_flag=True, help=(
            'Construct a reference image by max projecting imaging rounds and channels. Spots '
            'are found in this image and then measured across all images in the input stack.'
        )
    )
    @click.option(
        '--codebook', default=None, required=False, help=(
            'A spaceTx spec-compliant json file that describes a three dimensional tensor '
            'whose values are the expected intensity of a spot for each code in each imaging '
            'round and each color channel.'
        )
    )
    @click.pass_context
    def _cli(ctx, input, output, blobs_stack, reference_image_from_max_projection, codebook):
        print('Detecting Spots ...')
        ctx.obj = dict(
            component=SpotFinder,
            image_stack=ImageStack.from_path_or_url(input),
            output=output,
            blobs_stack=blobs_stack,
            reference_image_from_max_projection=reference_image_from_max_projection,
            codebook=None,
        )

        if codebook is not None:
            ctx.obj["codebook"] = Codebook.from_json(codebook)


SpotFinder._cli_register()



import json
import os
from typing import Any, BinaryIO, Callable, Dict, Mapping, Optional, Sequence, Tuple, Union

from slicedimage import (
    Collection,
    ImageFormat,
    Tile,
    TileSet,
    Writer,
)

from starfish.codebook.codebook import Codebook
from starfish.experiment.builder.orderediterator import join_dimension_sizes, ordered_iterator
from starfish.experiment.version import CURRENT_VERSION
from starfish.types import Coordinates, Indices
from .defaultproviders import RandomNoiseTile, tile_fetcher_factory
from .providers import FetchedTile, TileFetcher


AUX_IMAGE_NAMES = {
    'nuclei',
    'dots',
}
DEFAULT_DIMENSION_ORDER = (Indices.Z, Indices.ROUND, Indices.CH)


def _tile_opener(toc_path: str, tile: Tile, file_ext: str) -> BinaryIO:
    tile_basename = os.path.splitext(toc_path)[0]
    return open(
        "{}-Z{}-H{}-C{}.{}".format(
            tile_basename,
            tile.indices[Indices.Z],
            tile.indices[Indices.ROUND],
            tile.indices[Indices.CH],
            ImageFormat.TIFF.file_ext,
        ),
        "wb")


def _fov_path_generator(parent_toc_path: str, toc_name: str) -> str:
    toc_basename = os.path.splitext(os.path.basename(parent_toc_path))[0]
    return os.path.join(
        os.path.dirname(parent_toc_path),
        "{}-{}.json".format(toc_basename, toc_name),
    )


def build_image(
        fov_count: int, round_count: int, ch_count: int, z_count: int,
        image_fetcher: TileFetcher,
        default_shape: Optional[Tuple[int, int]]=None,
        dimension_order: Sequence[Indices]=DEFAULT_DIMENSION_ORDER,
) -> Collection:
    """
    Build and returns an image set with the following characteristics:

    Parameters
    ----------
    fov_count : int
        Number of fields of view in this image set.
    round_count : int
        Number for rounds in this image set.
    ch_count : int
        Number for channels in this image set.
    z_count : int
        Number of z-layers in this image set.
    image_fetcher : TileFetcher
        Instance of TileFetcher that provides the data for the tile.
    default_shape : Optional[Tuple[int, int]]
        Default shape of the individual tiles in this image set.
    dimension_order : Sequence[Indices]
        Ordering for which dimensions vary, in order of the slowest changing dimension to the
        fastest.  For instance, if the order is (ROUND, Z, CH) and each dimension has size 2, then
        the sequence is:
          (ROUND=0, CH=0, Z=0)
          (ROUND=0, CH=1, Z=0)
          (ROUND=0, CH=0, Z=1)
          (ROUND=0, CH=1, Z=1)
          (ROUND=1, CH=0, Z=0)
          (ROUND=1, CH=1, Z=0)
          (ROUND=1, CH=0, Z=1)
          (ROUND=1, CH=1, Z=1)
        (default = (Indices.Z, Indices.ROUND, Indices.CH))

    Returns
    -------
    The slicedimage collection representing the image.
    """
    dimension_sizes = join_dimension_sizes(
        dimension_order,
        size_for_round=round_count,
        size_for_ch=ch_count,
        size_for_z=z_count,
    )

    collection = Collection()
    for fov_ix in range(fov_count):
        fov_images = TileSet(
            [
                Coordinates.X,
                Coordinates.Y,
                Coordinates.Z,
                Indices.Z,
                Indices.ROUND,
                Indices.CH,
                Indices.X,
                Indices.Y,
            ],
            {Indices.ROUND: round_count, Indices.CH: ch_count, Indices.Z: z_count},
            default_shape,
            ImageFormat.TIFF,
        )

        for dimension_indices in ordered_iterator(dimension_sizes):
            image = image_fetcher.get_tile(
                fov_ix,
                dimension_indices[Indices.ROUND],
                dimension_indices[Indices.CH],
                dimension_indices[Indices.Z])
            tile = Tile(
                image.coordinates,
                {
                    Indices.Z: (dimension_indices[Indices.Z]),
                    Indices.ROUND: (dimension_indices[Indices.ROUND]),
                    Indices.CH: (dimension_indices[Indices.CH]),
                },
                image.shape,
                extras=image.extras,
            )
            tile.set_numpy_array_future(image.tile_data)
            fov_images.add_tile(tile)
        collection.add_partition("fov_{:03}".format(fov_ix), fov_images)
    return collection


def write_experiment_json(
        path: str,
        fov_count: int,
        primary_image_dimensions: Mapping[Union[str, Indices], int],
        aux_name_to_dimensions: Mapping[str, Mapping[Union[str, Indices], int]],
        primary_tile_fetcher: Optional[TileFetcher]=None,
        aux_tile_fetcher: Optional[Mapping[str, TileFetcher]]=None,
        postprocess_func: Optional[Callable[[dict], dict]]=None,
        default_shape: Optional[Tuple[int, int]]=None,
        dimension_order: Sequence[Indices]=(Indices.Z, Indices.ROUND, Indices.CH),
) -> None:
    """
    Build and returns a top-level experiment description with the following characteristics:

    Parameters
    ----------
    path : str
        Directory to write the files to.
    fov_count : int
        Number of fields of view in this experiment.
    primary_image_dimensions : Mapping[Union[str, Indices], int]
        Dictionary mapping dimension name to dimension size for the primary image.
    aux_name_to_dimensions : Mapping[str, Mapping[Union[str, Indices], int]]
        Dictionary mapping the auxiliary image type to dictionaries, which map from dimension name
        to dimension size.
    primary_tile_fetcher : Optional[TileFetcher]
        TileFetcher for primary images.  Set this if you want specific image data to be set for the
        primary images.  If not provided, the image data is set to random noise via
        :class:`RandomNoiseTileFetcher`.
    aux_tile_fetcher : Optional[Mapping[str, TileFetcher]]
        TileFetchers for auxiliary images.  Set this if you want specific image data to be set for
        one or more aux image types.  If not provided for any given aux image, the image data is
        set to random noise via :class:`RandomNoiseTileFetcher`.
    postprocess_func : Optional[Callable[[dict], dict]]
        If provided, this is called with the experiment document for any postprocessing.
        An example of this would be to add something to one of the top-level extras field.
        The callable should return what is to be written as the experiment document.
    default_shape : Optional[Tuple[int, int]] (default = None)
        Default shape for the tiles in this experiment.
    dimension_order : Sequence[Indices]
        Ordering for which dimensions vary, in order of the slowest changing dimension to the
        fastest.  For instance, if the order is (ROUND, Z, CH) and each dimension has size 2, then
        the sequence is:
          (ROUND=0, CH=0, Z=0)
          (ROUND=0, CH=1, Z=0)
          (ROUND=0, CH=0, Z=1)
          (ROUND=0, CH=1, Z=1)
          (ROUND=1, CH=0, Z=0)
          (ROUND=1, CH=1, Z=0)
          (ROUND=1, CH=0, Z=1)
          (ROUND=1, CH=1, Z=1)
        (default = (Indices.Z, Indices.ROUND, Indices.CH))
    """
    if primary_tile_fetcher is None:
        primary_tile_fetcher = tile_fetcher_factory(RandomNoiseTile)
    if aux_tile_fetcher is None:
        aux_tile_fetcher = {}
    if postprocess_func is None:
        postprocess_func = lambda doc: doc

    experiment_doc: Dict[str, Any] = {
        'version': str(CURRENT_VERSION),
        'images': {},
        'extras': {},
    }
    primary_image = build_image(
        fov_count,
        primary_image_dimensions[Indices.ROUND],
        primary_image_dimensions[Indices.CH],
        primary_image_dimensions[Indices.Z],
        primary_tile_fetcher,
        dimension_order=dimension_order,
        default_shape=default_shape,
    )
    Writer.write_to_path(
        primary_image,
        os.path.join(path, "primary_image.json"),
        pretty=True,
        partition_path_generator=_fov_path_generator,
        tile_opener=_tile_opener,
        tile_format=ImageFormat.TIFF,
    )
    experiment_doc['images']['primary'] = "primary_image.json"

    for aux_name, aux_dimensions in aux_name_to_dimensions.items():
        if aux_dimensions is None:
            continue
        auxiliary_image = build_image(
            fov_count,
            aux_dimensions[Indices.ROUND], aux_dimensions[Indices.CH], aux_dimensions[Indices.Z],
            aux_tile_fetcher.get(aux_name, tile_fetcher_factory(RandomNoiseTile)),
            dimension_order=dimension_order,
            default_shape=default_shape,
        )
        Writer.write_to_path(
            auxiliary_image,
            os.path.join(path, "{}.json".format(aux_name)),
            pretty=True,
            partition_path_generator=_fov_path_generator,
            tile_opener=_tile_opener,
            tile_format=ImageFormat.TIFF,
        )
        experiment_doc['images'][aux_name] = "{}.json".format(aux_name)

    experiment_doc["codebook"] = "codebook.json"
    with open(os.path.join(path, "experiment.json"), "w") as fh:
        json.dump(experiment_doc, fh, indent=4)

    codebook_array = [
        {
            "codeword": [
                {"r": 0, "c": 0, "v": 1},
            ],
            "target": "PLEASE_REPLACE_ME"
        },
    ]
    codebook = Codebook.from_code_array(codebook_array)
    codebook_json_filename = "codebook.json"
    codebook.to_json(os.path.join(path, codebook_json_filename))

    experiment_doc = postprocess_func(experiment_doc)



from functools import partial
from typing import Callable, Optional, Tuple, Union

import click
import numpy as np
from scipy.ndimage.filters import uniform_filter

from starfish.imagestack.imagestack import ImageStack
from starfish.types import Number
from ._base import FilterAlgorithmBase
from .util import (
    determine_axes_to_group_by, preserve_float_range, validate_and_broadcast_kernel_size
)


class MeanHighPass(FilterAlgorithmBase):

    def __init__(
            self, size: Union[Number, Tuple[Number]], is_volume: bool=False) -> None:
        """Mean high pass filter.

        The mean high pass filter reduces low spatial frequency features by subtracting a
        mean filtered image from the original image. The mean filter smooths an image by replacing
        each pixel's value with an average of the pixel values of the surrounding neighborhood.

        The mean filter is also known as a uniform or box filter.

        This is a pass through for the scipy.ndimage.filters.uniform_filter:
        https://docs.scipy.org/doc/scipy-0.19.0/reference/generated/scipy.ndimage.uniform_filter.html

        Parameters
        ----------
        size : Union[Number, Tuple[Number]]
            width of the kernel
        is_volume : bool
            If True, 3d (z, y, x) volumes will be filtered, otherwise, filter 2d tiles
            independently.

        """

        self.size = validate_and_broadcast_kernel_size(size, is_volume)
        self.is_volume = is_volume

    _DEFAULT_TESTING_PARAMETERS = {"size": 1}

    @staticmethod
    def _high_pass(image: np.ndarray, size: Number, rescale: bool=False) -> np.ndarray:
        """
        Applies a mean high pass filter to an image

        Parameters
        ----------
        image : numpy.ndarray[np.float32]
            2-d or 3-d image data
        size : Union[Number, Tuple[Number]]
            width of the kernel
        rescale : bool
            If true scales data by max value, if false clips max values to one

        Returns
        -------
        np.ndarray [np.float32]:
            Filtered image, same shape as input
        """

        blurred: np.ndarray = uniform_filter(image, size)

        filtered: np.ndarray = image - blurred
        filtered = preserve_float_range(filtered, rescale)

        return filtered

    def run(
            self, stack: ImageStack, in_place: bool=False, verbose: bool=False,
            n_processes: Optional[int]=None
    ) -> ImageStack:
        """Perform filtering of an image stack

        Parameters
        ----------
        stack : ImageStack
            Stack to be filtered.
        in_place : bool
            if True, process ImageStack in-place, otherwise return a new stack
        verbose : bool
            if True, report on filtering progress (default = False)
        n_processes : Optional[int]
            Number of parallel processes to devote to calculating the filter

        Returns
        -------
        ImageStack :
            If in-place is False, return the results of filter as a new stack.  Otherwise return the
            original stack.

        """
        group_by = determine_axes_to_group_by(self.is_volume)
        high_pass: Callable = partial(self._high_pass, size=self.size)
        result = stack.apply(
            high_pass,
            group_by=group_by, verbose=verbose, in_place=in_place, n_processes=n_processes
        )
        return result

    @staticmethod
    @click.command("MeanHighPass")
    @click.option(
        "--size", type=float, help="width of the kernel")
    @click.option(
        "--is-volume", is_flag=True,
        help="indicates that the image stack should be filtered in 3d")
    @click.pass_context
    def _cli(ctx, size, is_volume):
        ctx.obj["component"]._cli_run(ctx, MeanHighPass(size, is_volume))



import os
from tempfile import TemporaryDirectory

import numpy as np
import pytest

from starfish import ImageStack, IntensityTable
from starfish.types import Features


def test_to_mermaid_dataframe():
    """
    Creates a basic IntensityTable from an ImageStack and verifies that it can be dumped to disk
    as a DataFrame which MERmaid can load. Does not explicitly load the DataFrame in MERmaid.

    Verifies that the save function throws an error when target assignments are not present, which
    are required by MERmaid.
    """
    r, c, z, y, x = 1, 5, 2, 2, 5
    data = np.zeros(100, dtype=np.float32).reshape(r, c, z, y, x)
    image_stack = ImageStack.from_numpy_array(data)
    intensities = IntensityTable.from_image_stack(image_stack)

    # without a target assignment, should raise RuntimeError.
    with pytest.raises(RuntimeError):
        with TemporaryDirectory() as dir_:
            intensities.save_mermaid(os.path.join(dir_, 'test.csv.gz'))

    # assign targets
    intensities[Features.TARGET] = (Features.AXIS, np.random.choice(list('ABCD'), size=20))
    intensities[Features.DISTANCE] = (Features.AXIS, np.random.rand(20))
    with TemporaryDirectory() as dir_:
        intensities.save_mermaid(os.path.join(dir_, 'test.csv.gz'))



from typing import Mapping, Union

import xarray as xr

from starfish.types import Indices


def convert_to_indexers_dict(indexers) -> Mapping[str, Union[int, slice]]:
    """Converts a dict of indexers from Dict[Indices, (int/tuple)] to
    Dict[str, (int/slice)] so that it can be passed into standard xarray
    indexing methods

    Parameters
    ----------
    indexers : Dict[Indices, (int/tuple)]
            A dictionary of dim:index where index is the value or range to index the dimension

    """
    return_dict = {ind.value: slice(None, None) for ind in Indices}
    for key, value in indexers.items():
        if isinstance(value, tuple):
            value = slice(value[0], value[1])
        return_dict[key] = value
    return return_dict


def index_keep_dimensions(data: xr.DataArray, indexers: Mapping[str, Union[int, slice]]
                          ) -> xr.DataArray:
    """Takes an xarray and key to index it. Indexes then adds back in lost dimensions"""
    # store original dims
    original_dims = data.dims
    # index
    data = data.sel(indexers)
    # find missing dims
    missing_dims = set(original_dims) - set(data.dims)
    # Add back in missing dims
    data = data.expand_dims(tuple(missing_dims))
    # Reorder to correct format
    return data.transpose(*original_dims)



from typing import Optional

from starfish.imagestack.imagestack import ImageStack
from starfish.pipeline.algorithmbase import AlgorithmBase


class RegistrationAlgorithmBase(AlgorithmBase):
    def run(self, stack) -> Optional[ImageStack]:
        """Performs registration on the stack provided."""
        raise NotImplementedError()



from typing import Optional, Tuple

import matplotlib.axes
import matplotlib.pyplot as plt
import numpy as np

from starfish import IntensityTable
from starfish.types import Features
from .style import BACKGROUND_COLORMAP, TARGETS_COLORMAP
from .util import annotate_axis, dpi_correction, remove_axis_frame, set_default_arguments


# TODO use the name of the IntensityTable if it exists for a title
# TODO ambrosejcarr: currently the colormaps for decoded_image and intensities do not align.
# however, it should be possible to fix that.
def decoded_spots(
        intensities: IntensityTable=None,
        decoded_image: Optional[np.ndarray]=None,
        background_image: Optional[np.ndarray]=None,
        background_shape: Optional[Tuple[int, int]]=None,
        ax: Optional[matplotlib.axes.Axes]=None,
        background_kwargs: Optional[dict]=None,
        spots_kwargs: Optional[dict]=None,
        decoded_image_kwargs: Optional[dict]=None,
) -> matplotlib.axes.Axes:
    """Plot decoded spots with a variety of options for background image

    For plotting spots, provide either:

    1. ``intensities``, in which case spots will be plotted based on the estimated
       radius and position for each feature, or
    2. ``decoded_image``, in which case pixels will be colored by their decoded feature

    For plotting background, provide either:

    1. ``background_image``, which will plot the image for use as a background. Commonly this could
       include a nuclei or dots image.
    2. ``background_shape``, which will be used to construct a black background against which spots
       will be plotted. If decoded_image is provided as the spots parameter, then background
       arguments can be omitted, since the shape of the background can be inferred from the
       decoded_image.

    Parameters
    ----------
    intensities : Optional[IntensityTable]
        contains spots to plot
    decoded_image : Optional[np.ndarray]
        image, where values are coded to represent features
    background_image : Optional[np.ndarray]
        an image atop which spots or decoded pixels should be plotted
    background_shape : Optional[Tuple[int, int]]
        the shape for a blank background. Required when background_image is not provided and
        intensities are passed.
    ax : Optional[matplotlib.axes.Axes]
    background_kwargs, spots_kwargs, decoded_image_kwargs : Optional[dict]
        Keyword arguments to pass to plotting functions for background, spots, and decoded_image

    Returns
    -------
    matplotlib.axes.Axes :
        The axis the plot was constructed in.

    """

    # check that enough information on the image extent has been provided
    if all(p is None for p in (background_image, background_shape, decoded_image)):
        raise ValueError(
            'One of background_image, background_shape, or decoded_image must be provided to '
            'set the pixel size of the plot.'
        )

    if intensities is None and decoded_image is None:
        raise ValueError(
            'One of intensities or decoded_image must be provided to visualize spots '
        )

    # if needed, get an axis to plot in
    if ax is None:
        ax = plt.gca()

    # set backgrund plotting arguments
    background_kwargs = set_default_arguments(background_kwargs, cmap=BACKGROUND_COLORMAP)

    # determine what the background should be
    if background_image is None:
        if decoded_image is not None:
            background_image = np.zeros_like(decoded_image)
        else:
            background_image = np.zeros(background_shape)

    # plot the background
    ax.imshow(background_image, **background_kwargs)

    if intensities is not None:
        all_targets = set(np.unique(intensities[Features.TARGET])) - {'None'}
        color_map = dict(zip(all_targets, np.arange(len(all_targets))))

        x = intensities.x[intensities[Features.TARGET] != 'None'].values
        y = intensities.y[intensities[Features.TARGET] != 'None'].values

        targets = intensities[Features.TARGET][intensities[Features.TARGET] != 'None']
        colors = [color_map[t] for t in targets.values]

        # size in matplotlib is a bit involved. First, get the radii
        size = intensities.radius[intensities[Features.TARGET] != 'None'].values
        # make a DPI correction
        size = size * dpi_correction()
        # scatterplot produces dots of size equal to sqrt(s), so square to get correct radius
        size = size ** 2

        # set spots visualization parameters
        spots_kwargs = set_default_arguments(
            spots_kwargs,
            cmap=TARGETS_COLORMAP,
            alpha=0.6
        )

        ax.scatter(x, y, s=size, c=colors, **spots_kwargs)

    if decoded_image is not None:

        # set decoded parameters
        decoded_image_kwargs = set_default_arguments(
            decoded_image_kwargs,
            alpha=0.5,
            cmap=TARGETS_COLORMAP
        )

        masked_image = np.ma.masked_equal(decoded_image, 0)
        ax.imshow(masked_image, interpolation='none', **decoded_image_kwargs)

    remove_axis_frame(ax)
    annotate_axis(ax, title='Decoded Image')
    return ax



from functools import partial
from multiprocessing import Pool
from typing import Dict, List, NamedTuple, Optional, Tuple

import numpy as np
import pandas as pd
from skimage.measure import label, regionprops
from skimage.measure._regionprops import _RegionProperties
from tqdm import tqdm

from starfish.intensity_table.intensity_table import IntensityTable
from starfish.types import Features, Indices, Number, SpotAttributes


class ConnectedComponentDecodingResult(NamedTuple):
    region_properties: List
    label_image: np.ndarray
    decoded_image: np.ndarray


class TargetsMap:

    def __init__(self, targets: np.ndarray) -> None:
        """
        Creates an invertible mapping between string names of Codebook targets and integer IDs
        that can be interpreted by skimage.measure to decode an image.

        Parameters
        ----------
        targets : np.ndarray
            array of string target IDs

        """
        unique_targets = set(targets) - {'nan'}
        sorted_targets = sorted(unique_targets)
        self._int_to_target = dict(zip(range(1, np.iinfo(np.int).max), sorted_targets))
        self._int_to_target[0] = 'nan'
        self._target_to_int = {v: k for (k, v) in self._int_to_target.items()}

    def targets_as_int(self, targets: np.ndarray) -> np.ndarray:
        """Transform an array of targets into their integer representation.

        Parameters
        ----------
        targets : np.ndarray['U']
            array of string targets to be transformed into integer IDs

        Returns
        -------
        np.ndarray[int] :
            array of targets represented by their integer IDs

        """
        return np.array([self._target_to_int[v] for v in targets])

    def targets_as_str(self, targets: np.ndarray) -> np.ndarray:
        """Transform an array of integer IDs into their corresponding string target names.

        Parameters
        ----------
        targets : np.ndarray[int]
            array of int targets to be transformed into string names

        Returns
        -------
        np.ndarray['U']
            array of unicode-encoded target names

        """
        return np.array([self._int_to_target[v] for v in targets])

    def target_as_str(self, integer_target: int) -> np.ndarray:
        return self._int_to_target[integer_target]


class CombineAdjacentFeatures:

    def __init__(
            self,
            min_area: Number,
            max_area: Number,
            connectivity: int=2,
            mask_filtered_features: bool=True
    ) -> None:
        """Combines pixel-wise adjacent features into single larger features using skimage.measure

        Parameters
        ----------
        min_area : Number
            Combined features with area below this value are marked as failing filters
        max_area : Number
            Combined features with area above this value are marked as failing filters
        connectivity : int
            Maximum number of orthogonal hops to consider a pixel/voxel as a neighbor. See
            http://scikit-image.org/docs/dev/api/skimage.measure.html#skimage.measure.label for more
            detail. Default = 2.
        mask_filtered_features : bool
            If True, sets all pixels that are failing filters applied prior to this function equal
            to zero, the background value for skimage.measure.label

        """
        self._min_area = min_area
        self._max_area = max_area
        self._connectivity = connectivity
        self._mask_filtered = mask_filtered_features

    @staticmethod
    def _intensities_to_decoded_image(
            intensities: IntensityTable,
            target_map: TargetsMap,
            mask_filtered_features: bool=True
    ) -> np.ndarray:
        """
        Construct an image where each pixel corresponds to its decoded target, mapped to a unique
        integer ID

        Parameters
        ----------
        intensities : IntensityTable
            Decoded intensities
        target_map : TargetsMap
            Mapping between string target names and integer target IDs
        mask_filtered_features : bool
            If true, all features that fail filters are mapped to zero, which is considered
            'background' and will not decode to a target (default = True).

        Returns
        -------
        np.ndarray[int]
            Image whose pixels are coded as the targets that the ImageStack decoded to at each
            position.

        """
        # reverses the linearization that was used to transform an ImageStack into an IntensityTable
        max_x = intensities[Indices.X.value].values.max() + 1
        max_y = intensities[Indices.Y.value].values.max() + 1
        max_z = intensities[Indices.Z.value].values.max() + 1

        int_targets = target_map.targets_as_int(intensities[Features.TARGET].values)
        if mask_filtered_features:
            fails_filters = np.where(~intensities[Features.PASSES_THRESHOLDS])[0]
            int_targets[fails_filters] = 0

        decoded_image: np.ndarray = int_targets.reshape((max_z, max_y, max_x))
        return decoded_image

    @staticmethod
    def _calculate_mean_pixel_traces(
            label_image: np.ndarray,
            intensities: IntensityTable,
    ) -> IntensityTable:
        """
        For all pixels that contribute to a connected component, calculate the mean value for
        each (ch, round), producing an average "trace" of a feature across the imaging experiment

        Parameters
        ----------
        label_image : np.ndarray
            An image where all pixels of a connected component share the same integer ID
        intensities : IntensityTable
            decoded intensities

        Returns
        -------
        IntensityTable :
            an IntensityTable where the number of features equals the number of connected components
            and the intensities of each each feature is its mean trace.

        """
        pixel_labels = label_image.reshape(-1)
        intensities['spot_id'] = (Features.AXIS, pixel_labels)
        mean_pixel_traces = intensities.groupby('spot_id').mean(Features.AXIS)
        mean_distances = intensities[Features.DISTANCE].groupby('spot_id').mean(Features.AXIS)
        mean_pixel_traces[Features.DISTANCE] = (
            'spot_id',
            np.ravel(mean_distances)
        )

        # the 0th pixel trace corresponds to background. If present, drop it.
        try:
            mean_pixel_traces = mean_pixel_traces.drop(0, dim='spot_id')
        except KeyError:
            pass

        return mean_pixel_traces

    @staticmethod
    def _single_spot_attributes(
            spot_property: _RegionProperties,
            decoded_image: np.ndarray,
            target_map: TargetsMap,
            min_area: Number,
            max_area: Number,
    ) -> Tuple[Dict[str, int], int]:
        """
        Calculate starfish SpotAttributes from the RegionProperties of a connected component
        feature.

        Parameters
        ----------
        spot_property: _RegionProperties
            Properties of the connected component. Output of skimage.measure.regionprops
        decoded_image : np.ndarray
            Image whose pixels correspond to the targets that the given position in the ImageStack
            decodes to.
        target_map : TargetsMap
            Unique mapping between string target names and int target IDs.
        min_area :
            Combined features with area below this value are marked as failing filters
        max_area : Number
            Combined features with area above this value are marked as failing filters

        Returns
        -------
        Dict[str, Number] :
            spot attribute dictionary for this connected component, containing the x, y, z position,
            target name (str) and feature radius.
        int :
            1 if spot passes size filters, zero otherwise.

        """
        # because of the above skimage issue, we need to support both 2d and 3d properties
        if len(spot_property.centroid) == 3:
            spot_attrs = {
                'z': int(spot_property.centroid[0]),
                'y': int(spot_property.centroid[1]),
                'x': int(spot_property.centroid[2])
            }
        else:  # data is 2d
            spot_attrs = {
                'z': 0,
                'y': int(spot_property.centroid[0]),
                'x': int(spot_property.centroid[1])
            }

        # we're back to 3d or fake-3d here
        target_index = decoded_image[spot_attrs['z'], spot_attrs['y'], spot_attrs['x']]
        spot_attrs[Features.TARGET] = target_map.target_as_str(target_index)
        spot_attrs[Features.SPOT_RADIUS] = spot_property.equivalent_diameter / 2

        # filter intensities for which radius is too small
        passes_area_filter = 1 if min_area <= spot_property.area < max_area else 0
        return spot_attrs, passes_area_filter

    def _create_spot_attributes(
            self,
            region_properties: List[_RegionProperties],
            decoded_image: np.ndarray,
            target_map: TargetsMap,
            n_processes: Optional[int]=None
    ) -> Tuple[SpotAttributes, np.ndarray]:
        """

        Parameters
        ----------
        region_properties : List[_RegionProperties]
            Properties of the each connected component. Output of skimage.measure.regionprops
        decoded_image : np.ndarray
            Image whose pixels correspond to the targets that the given position in the ImageStack
            decodes to.
        target_map : TargetsMap
            Unique mapping between string target names and int target IDs.
        n_processes : Optional[int]=None
            number of processes to devote to measuring spot properties. If None, defaults to the
            result of os.nproc()

        Returns
        -------
        pd.DataFrame :
            DataFrame containing x, y, z, radius, and target name for each connected component
            feature.
        np.ndarray[bool] :
            An array with length equal to the number of features. If zero, indicates that a feature
            has failed area filters.
        """
        pool = Pool(n_processes)
        mapfunc = pool.map
        applyfunc = partial(
            self._single_spot_attributes,
            decoded_image=decoded_image,
            target_map=target_map,
            min_area=self._min_area,
            max_area=self._max_area
        )

        iterable = tqdm(region_properties)
        results = mapfunc(applyfunc, iterable)
        spot_attrs, passes_area_filter = zip(*results)

        # update passes filter
        passes_filter = np.array(passes_area_filter, dtype=np.bool)

        spot_attributes = SpotAttributes(pd.DataFrame.from_records(spot_attrs))
        return spot_attributes, passes_filter

    def run(
            self, intensities: IntensityTable
    ) -> Tuple[IntensityTable, ConnectedComponentDecodingResult]:
        """
        Execute the combine_adjacent_features method on an IntensityTable containing pixel
        intensities

        Parameters
        ----------
        intensities : IntensityTable
            Pixel intensities of an imaging experiment

        Returns
        -------
        IntensityTable :
            Table whose features comprise sets of adjacent pixels that decoded to the same target
        ConnectedComponentDecodingResult :
            NamedTuple containing :
                region_properties :
                    the properties of each connected component, in the same order as the
                    IntensityTable
                label_image : np.ndarray
                    An image where all pixels of a connected component share the same integer ID
                decoded_image : np.ndarray
                    Image whose pixels correspond to the targets that the given position in the
                    ImageStack decodes to.

        """

        # map target molecules to integers so they can be reshaped into an image that can
        # be subjected to a connected-component algorithm to find adjacent pixels with the
        # same targets
        targets = intensities[Features.TARGET].values
        target_map = TargetsMap(targets)

        # create the decoded_image
        decoded_image = self._intensities_to_decoded_image(
            intensities,
            target_map,
            self._mask_filtered,
        )

        # label the decoded image to extract connected component features
        label_image: np.ndarray = label(decoded_image, connectivity=self._connectivity)

        # calculate properties of each feature
        props: List = regionprops(np.squeeze(label_image))

        # calculate mean intensities across the pixels of each feature
        mean_pixel_traces = self._calculate_mean_pixel_traces(
            label_image,
            intensities,
        )

        # Create SpotAttributes and determine feature filtering outcomes
        spot_attributes, passes_filter = self._create_spot_attributes(
            props,
            decoded_image,
            target_map,
        )

        # augment the SpotAttributes with filtering results and distances from nearest codes
        spot_attributes.data[Features.DISTANCE] = mean_pixel_traces[Features.DISTANCE]
        spot_attributes.data[Features.PASSES_THRESHOLDS] = passes_filter

        # create new indexes for the output IntensityTable
        channel_index = mean_pixel_traces.indexes[Indices.CH]
        round_index = mean_pixel_traces.indexes[Indices.ROUND]
        coords = IntensityTable._build_xarray_coords(spot_attributes, channel_index, round_index)

        # create the output IntensityTable
        dims = (Features.AXIS, Indices.CH.value, Indices.ROUND.value)
        intensity_table = IntensityTable(
            data=mean_pixel_traces, coords=coords, dims=dims
        )

        # combine the various non-IntensityTable results into a NamedTuple before returning
        ccdr = ConnectedComponentDecodingResult(props, label_image, decoded_image)

        return intensity_table, ccdr



from functools import partial
from typing import Optional

import click
import numpy as np

from starfish.imagestack.imagestack import ImageStack
from ._base import FilterAlgorithmBase
from .util import determine_axes_to_group_by, preserve_float_range


class ScaleByPercentile(FilterAlgorithmBase):

    def __init__(self, p: int=0, is_volume: bool=False) -> None:
        """Image scaling filter

        Parameters
        ----------
        p : int
            each image in the stack is scaled by this percentile. must be in [0, 100]
        is_volume : bool
            If True, 3d (z, y, x) volumes will be filtered. By default, filter 2-d (y, x) tiles
        """
        self.p = p
        self.is_volume = is_volume

    _DEFAULT_TESTING_PARAMETERS = {"p": 0}

    @staticmethod
    def _scale(image: np.ndarray, p: int) -> np.ndarray:
        """Clip values of img below and above percentiles p_min and p_max

        Parameters
        ----------
        image : np.ndarray
            image to be scaled

        p : int
            each image in the stack is scaled by this percentile. must be in [0, 100]

        Notes
        -----
        - Setting p to 100 scales the image by it's maximum value
        - No shifting or transformation to adjust dynamic range is done after scaling

        Returns
        -------
        np.ndarray :
          Numpy array of same shape as img

        """
        v = np.percentile(image, p)

        image = image / v
        image = preserve_float_range(image)

        return image

    def run(
            self, stack: ImageStack, in_place: bool=False, verbose: bool=False,
            n_processes: Optional[int]=None
    ) -> ImageStack:
        """Perform filtering of an image stack

        Parameters
        ----------
        stack : ImageStack
            Stack to be filtered.
        in_place : bool
            if True, process ImageStack in-place, otherwise return a new stack
        verbose : bool
            If True, report on the percentage completed (default = False) during processing
        n_processes : Optional[int]
            Number of parallel processes to devote to calculating the filter

        Returns
        -------
        ImageStack :
            If in-place is False, return the results of filter as a new stack.  Otherwise return the
            original stack.

        """
        group_by = determine_axes_to_group_by(self.is_volume)
        clip = partial(self._scale, p=self.p)
        result = stack.apply(
            clip,
            group_by=group_by, verbose=verbose, in_place=in_place, n_processes=n_processes
        )
        return result

    @staticmethod
    @click.command("ScaleByPercentile")
    @click.option(
        "--p", default=100, type=int, help="scale images by this percentile")
    @click.option(  # FIXME: was this intentionally missed?
        "--is-volume", is_flag=True, help="filter 3D volumes")
    @click.pass_context
    def _cli(ctx, p, is_volume):
        ctx.obj["component"]._cli_run(ctx, ScaleByPercentile(p, is_volume))



import numpy as np
import pandas as pd
import pytest

from starfish import Codebook, IntensityTable
from starfish.types import Features, Indices, SpotAttributes


def intensity_table_factory() -> IntensityTable:
    """IntensityTable with a single feature that was measured over 2 channels and 2 rounds."""

    intensities = np.array(
        [[[0, 3],
          [4, 0]]],
        dtype=float
    )
    spot_attribute_data = pd.DataFrame(
        data=[0, 0, 0, 1],
        index=[Indices.Z, Indices.Y, Indices.X, Features.SPOT_RADIUS]
    ).T
    spot_attributes = SpotAttributes(spot_attribute_data)

    intensity_table = IntensityTable.from_spot_data(intensities, spot_attributes)
    return intensity_table


def codebook_factory() -> Codebook:
    """
    Codebook with two codewords describing an experiment with three channels and two imaging rounds.
    Both codes have two "on" channels.
    """
    codebook_array = [
        {
            Features.CODEWORD: [
                {Indices.ROUND.value: 0, Indices.CH.value: 0, Features.CODE_VALUE: 1},
                {Indices.ROUND.value: 1, Indices.CH.value: 1, Features.CODE_VALUE: 1}
            ],
            Features.TARGET: "GENE_A"
        },
        {
            Features.CODEWORD: [
                {Indices.ROUND.value: 0, Indices.CH.value: 2, Features.CODE_VALUE: 1},
                {Indices.ROUND.value: 1, Indices.CH.value: 1, Features.CODE_VALUE: 1}
            ],
            Features.TARGET: "GENE_B"
        },
    ]
    return Codebook.from_code_array(codebook_array)


@pytest.mark.parametrize('norm_order, expected_size', [
    (1, 2),
    (2, np.sqrt(2))
])
def test_normalize_codes(norm_order, expected_size):
    """
    Create a simple codebook with two features, each with two "on" sites. For these simple cases,
    we can easily calculate the expected value of the norms. (2-norm: sqrt 2, 1-norm: 2), Verify
    that the output of the functions correspond to these expectations.
    """
    codebook = codebook_factory()
    normed_codebook, norms = Codebook._normalize_features(codebook, norm_order)
    assert np.all(norms == expected_size)

    # each code should still have only two non-zero values
    assert np.all(
        normed_codebook.groupby(Features.TARGET).apply(lambda x: np.sum(x != 0)) == 2
    )

    # each non-zero value should be equal to 1 / expected_size of the norm. There are two non-zero
    # values and so the sum of the code should be (1 / expected_size) * 2
    assert np.all(
        normed_codebook.sum((Indices.CH.value, Indices.ROUND.value)) == (1 / expected_size) * 2
    )


@pytest.mark.parametrize('norm_order, expected_size', [
    (1, 7),
    (2, 5)
])
def test_normalize_intensities(norm_order, expected_size):
    """
    Create a slightly less simple IntensityTable with one "on" tile per feature,
    we can again calculate the expected value of the norms. Verify
    that the output of the functions correspond to these expectations.
    """
    intensity_table = intensity_table_factory()
    normed_intensities, norms = Codebook._normalize_features(intensity_table, norm_order)

    assert np.all(norms == expected_size)

    # each feature should still have only two non-zero values
    assert np.all(
        normed_intensities.groupby(Features.AXIS).apply(lambda x: np.sum(x != 0)) == 2
    )

    # each non-zero value should be equal to 1 / expected_size of the norm.
    assert np.all(normed_intensities == intensity_table / norms)


# TODO I think the outcome of this test should be NaNs -- it's too sensitive to boundary conditions
# 0/0/0/0 vs 0/0/0/1e-30 give VERY different results
@pytest.mark.skip('Test is wrong, needs to be revisited')
@pytest.mark.parametrize('norm_order, expected_value', [
    (1, 1 / 4),
    (2, 1 / 8)
])
def test_all_blank_features_yield_non_zero_but_equal_normalized_values(norm_order, expected_value):

    intensity_table = intensity_table_factory()

    # zero-out all the values in the IntensityTable
    intensity_table.values = np.zeros(4).reshape(1, 2, 2)
    normed_intensities, norms = Codebook._normalize_features(intensity_table, norm_order)

    # todo norms here are zero, which seems like the right answer!
    assert norms == 0

    assert np.all(normed_intensities == expected_value)



"""
Tests for CombineAdjacentFeatures._calculate_mean_pixel_traces method
"""

from typing import Tuple

import numpy as np
from skimage.measure import label

from starfish import ImageStack, IntensityTable
from starfish.spots._detector.combine_adjacent_features import CombineAdjacentFeatures
from starfish.types import Features


# TODO ambrosejcarr: make sure these arrays contain background pixels to catch  one-off errors
def labeled_intensities_factory() -> Tuple[IntensityTable, np.ndarray, np.ndarray]:
    """
    Create a decoded IntensityTable with distance scores, and a corresponding label_image and
    decoded_image.
    """
    data = np.array(
        [[[[0., 0.], [.1, .1]],  # ch 1
          [[.5, .5], [.2, .3]]],
         [[[.1, .1], [0, 0]],  # ch 2, x & y are reversed
          [[.2, .3], [.5, .5]]]],
        dtype=np.float32
    )
    image_stack = ImageStack.from_numpy_array(data.reshape(1, 2, 2, 2, 2))
    intensity_table = IntensityTable.from_image_stack(image_stack)
    intensity_table[Features.DISTANCE] = (Features.AXIS, np.zeros(intensity_table.shape[0]))
    label_image = np.array(
        [[[1, 2], [3, 4]],
         [[1, 2], [3, 4]]]
    )
    decoded_image = np.array(
        [[[5, 4], [3, 2]],
         [[5, 4], [3, 2]]]
    )

    # verify that the listed label image is what would be created by the function we use in the
    # code
    assert np.array_equal(label(decoded_image), label_image)

    return intensity_table, label_image, decoded_image


# TODO ambrosejcarr: increase test fixture complexity to verify this works over multiple rounds
def test_calculate_mean_pixel_traces():
    """
    Test that calculate_mean_pixel_traces matches the mean trace produced from the ndarray data
    used to construct the testing fixtures.
    """
    intensity_table, label_image, _ = labeled_intensities_factory()

    passes_filter = np.ones(4)
    mean_pixel_traces = CombineAdjacentFeatures._calculate_mean_pixel_traces(
        label_image,
        intensity_table,
    )

    # evaluate the mean pixel traces, we have 4 different spot ids
    assert np.unique(label_image).shape[0] == mean_pixel_traces.shape[0]

    # there should be two channels and 1 round
    assert np.array_equal(intensity_table.shape[1:], mean_pixel_traces.shape[1:])

    # values can be calculated from the simple example, and should match the mean pixel traces
    expected_values = np.array(
        [[[.25],
          [.15]],
         [[.25],
          [.2]],
         [[.15],
          [.25]],
         [[.2],
          [.25]]],
        dtype=np.float32
    )
    assert np.allclose(expected_values, mean_pixel_traces.values)

    # no values should be filtered, as all spots decoded
    assert np.all(passes_filter)



from typing import Optional, Tuple

import click
import numpy as np
import scipy.ndimage.measurements as spm
from scipy.ndimage import distance_transform_edt
from showit import image
from skimage.feature import peak_local_max
from skimage.morphology import watershed

from starfish.image._filter.util import bin_open, bin_thresh
from starfish.imagestack.imagestack import ImageStack
from starfish.types import Indices, Number
from ._base import SegmentationAlgorithmBase


class Watershed(SegmentationAlgorithmBase):

    def __init__(
        self,
        nuclei_threshold: Number,
        input_threshold: Number,
        min_distance: int,
        **kwargs
    ) -> None:
        """Implements watershed segmentation of cells.

        Algorithm is seeded by nuclei image. Binary segmentation mask is computed from a maximum
        projection of spots across C and R, which is subsequently thresholded.

        This function wraps skimage watershed.

        Parameters
        ----------
        nuclei_threshold : Number
            threshold to apply to nuclei image
        input_threshold : Number
            threshold to apply to stain image
        min_distance : int
            minimum distance before object centers in a provided nuclei image are considered single
            nuclei

        See Also
        --------
        Watershed: http://scikit-image.org/docs/dev/auto_examples/segmentation/plot_watershed.html

        """
        self.nuclei_threshold = nuclei_threshold
        self.input_threshold = input_threshold
        self.min_distance = min_distance
        self._segmentation_instance: Optional[_WatershedSegmenter] = None

    def run(self, primary_images: ImageStack, nuclei: ImageStack) -> np.ndarray:
        """Segments nuclei in 2-d using a nuclei ImageStack

        Primary images are used to expand the nuclear mask, but only in cases where there are
        densely detected points surrounding the nuclei.

        Parameters
        ----------
        primary_images : ImageStack
            contains primary image data
        nuclei : ImageStack
            contains nuclei image data

        Returns
        -------
        np.ndarray :
            label image where each cell is labeled by a different positive integer value. 0
            implies that a pixel is not part of a cell.
        """

        # create a 'stain' for segmentation
        mp = primary_images.max_proj(Indices.CH, Indices.Z)
        mp_numpy = mp._squeezed_numpy(Indices.CH, Indices.Z)
        stain = np.mean(mp_numpy, axis=0)
        stain = stain / stain.max()

        # TODO make these parameterizable or determine whether they are useful or not
        size_lim = (10, 10000)
        disk_size_markers = None
        disk_size_mask = None

        nuclei_mp = nuclei.max_proj(Indices.ROUND, Indices.CH, Indices.Z)
        nuclei__mp_numpy = nuclei_mp._squeezed_numpy(Indices.ROUND, Indices.CH, Indices.Z)
        self._segmentation_instance = _WatershedSegmenter(nuclei__mp_numpy, stain)
        label_image = self._segmentation_instance.segment(
            self.nuclei_threshold, self.input_threshold, size_lim, disk_size_markers,
            disk_size_mask, self.min_distance
        )

        return label_image

    def show(self, figsize: Tuple[int, int]=(10, 10)) -> None:
        if isinstance(self._segmentation_instance, _WatershedSegmenter):
            self._segmentation_instance.show(figsize=figsize)
        else:
            raise RuntimeError('Run segmentation before attempting to show results.')

    @staticmethod
    @click.command("Watershed")
    @click.option(
        "--nuclei-threshold", default=.16, type=float, help="Nuclei threshold")
    @click.option(
        "--input-threshold", default=.22, type=float, help="Input threshold")
    @click.option(
        "--min-distance", default=57, type=int, help="Minimum distance between cells")
    @click.pass_context
    def _cli(ctx, nuclei_threshold, input_threshold, min_distance):
        ctx.obj["component"]._cli_run(
            ctx, Watershed(nuclei_threshold, input_threshold, min_distance))


class _WatershedSegmenter:
    def __init__(self, nuclei_img: np.ndarray, stain_img: np.ndarray) -> None:
        """Implements watershed segmentation of cells seeded from a nuclei image

        Algorithm is seeded by a nuclei image. Binary segmentation mask is computed from a maximum
        projection of spots across C and R, which is subsequently thresholded.

        Parameters
        ----------
        nuclei_img : np.ndarray[np.float32]
            nuclei image
        stain_img : np.ndarray[np.float32]
            stain image
        """
        self.nuclei = nuclei_img / nuclei_img.max()
        self.stain = stain_img / stain_img.max()

        self.nuclei_thresholded: Optional[np.ndarray] = None  # dtype: bool
        self.markers = None
        self.num_cells: Optional[int] = None
        self.mask = None
        self.segmented = None

    def segment(
            self,
            nuclei_thresh: Number,
            stain_thresh: Number,
            size_lim: Tuple[int, int],
            disk_size_markers: Optional[int]=None,  # TODO ambrosejcarr what is this doing?
            disk_size_mask: Optional[int]=None,  # TODO ambrosejcarr what is this doing?
            min_dist: Optional[Number]=None
    ) -> np.ndarray:
        """Execute watershed cell segmentation.

        Parameters
        ----------
        nuclei_thresh, stain_thresh : Number
            Threshold for the nuclei and stain images. All pixels with intensities above this size
            will be considered part of the objects in the respective images
        size_lim : Tuple[int, int]
            min and max allowable size for nuclei objects in the nuclei_image
        disk_size_markers : Optional[int]
            if provided ...
        disk_size_mask : Optional[int]
            if provided ...
        min_dist : Optional[int]
            if provided, nuclei within this distance of each other are combined into single
            objects.

        Returns
        -------
        np.ndarray[int32] :
            label image with same size and shape as self.nuclei_img
        """
        min_allowed_size, max_allowed_size = size_lim
        self.nuclei_thresholded = self.filter_nuclei(nuclei_thresh, disk_size_markers)
        self.markers, self.num_cells = self.label_nuclei(
            self.nuclei_thresholded,
            min_allowed_size, max_allowed_size, min_dist
        )
        self.mask = self.watershed_mask(stain_thresh, self.markers, disk_size_mask)
        self.segmented = self.watershed(self.markers, self.mask)
        return self.segmented

    def filter_nuclei(self, nuclei_thresh: float, disk_size: Optional[int]) -> np.ndarray:
        """Threshold the nuclei image at nuclei_thresh.

        Parameters
        ----------
        nuclei_thresh : float
            Threshold the nuclei image at this value
        disk_size : int
            if passed, execute a binary opening of the filtered image

        Returns
        -------
        np.ndarray[bool] :
            thresholded image
        """
        nuclei_filt = bin_thresh(self.nuclei, nuclei_thresh)
        if disk_size is not None:
            nuclei_filt = bin_open(nuclei_filt, disk_size)
        return nuclei_filt

    def label_nuclei(
        self,
        nuclei_thresholded: np.ndarray,
        min_allowed_size: int,
        max_allowed_size: int,
        min_dist: Optional[Number]=None
    ) -> Tuple[np.ndarray, int]:
        """Construct a labeled nuclei image, which will be combined with the point cloud to seed
        the watershed

        Parameters
        ----------
        nuclei_thresholded : np.ndarray
            thresholded nuclei image
        min_allowed_size : int
            minimum allowable thresholded nuclei size
        max_allowed_size : int
            maximum allowable nuclei size

        Returns
        -------
        np.ndarray :
            labeled nuclei, excluding those whose size is outside the area boundaries

        """

        # label thresholded nuclei image
        if min_dist is None:
            markers, num_objs = spm.label(nuclei_thresholded)
        else:
            markers, num_objs = self._unclump(min_dist)

        # TODO dganguli: does it really make sense to assume a square area?
        min_allowed_area = min_allowed_size ** 2
        max_allowed_area = max_allowed_size ** 2

        # spm.sum sums the values of an array by label. This counts the pixels in each object
        areas = spm.sum(
            np.ones(nuclei_thresholded.shape),
            markers,
            np.array(range(0, num_objs + 1), dtype=np.int32)
        )

        # each label value is replaced by its area
        area_image = areas[markers]

        # areas are used to mask values that are outside the allowable sizes
        markers[area_image <= min_allowed_area] = 0
        markers[area_image >= max_allowed_area] = 0

        # re-label the image with sequential integers, accounting for exclusion based on size
        markers_reduced, num_objs = self.relabel_image(markers)

        return markers_reduced, num_objs

    def _unclump(self, min_dist: Number) -> Tuple[np.ndarray, int]:
        """
        Run watershed on the thresholded basin image, restricted to basins at least min_dist apart

        Functionally, this reproduces the thresholded nuclei image with overlapping nuclei merged.

        Parameters
        ----------
        min_dist : int
            minimum distance between watershed basins

        """
        im: np.ndarray = self.nuclei_thresholded

        # calculates the distance of every pixel to the nearest background (0) point
        distance: np.ndarray = distance_transform_edt(im)  # dtype: np.float64

        # boolean array marking local maxima, excluding any maxima within min_dist
        local_maxi: np.ndarray = peak_local_max(
            distance, labels=im, indices=False, min_distance=min_dist
        )

        # label the maxima for watershed
        markers, num_objs = spm.label(local_maxi)

        # run watershed, using the distances in the thresholded image as basins.
        # Uses the original image as a mask, preventing any background pixels from being labeled
        labels_ws: np.ndarray = watershed(-distance, markers, mask=im)
        return labels_ws, num_objs

    def watershed_mask(self, stain_thresh: Number, markers: np.ndarray, disk_size: Optional[int]):
        """Create a watershed mask that is the union of the spot intensities above stain_thresh and
        a marker image generated from nuclei

        Parameters
        ----------
        stain_thresh : Number
            threshold to apply to the stain image
        markers : np.ndarray[bool]
            markers for the stain_image
        disk_size : Optional[int]
            if provided, execute a morphological opening operation over the thresholded stain image

        Returns
        -------
        np.ndarray[bool] :
            thresholded stain image

        """
        st = self.stain >= stain_thresh
        watershed_mask: np.ndarray = np.logical_or(st, markers > 0)  # dtype bool
        if disk_size is not None:
            watershed_mask = bin_open(watershed_mask, disk_size)
        return watershed_mask

    def watershed(self, markers: np.ndarray, watershed_mask: np.ndarray) -> np.ndarray:
        """Run watershed on the thresholded primary_images max projection

        Parameters
        ----------
        markers : np.ndarray[np.int64]
            an array marking the basins with the values to be assigned in the label matrix.
            Zero means not a marker.
        watershed_mask : np.ndarray[bool]
            Mask array. only points at which mask == True will be labeled in the output.

        Returns
        -------
        np.ndarray[np.int32] :
            labeled image, each segment has a unique integer value
        """
        img = 1 - self.stain

        res = watershed(image=img,
                        markers=markers,
                        connectivity=np.ones((3, 3), bool),
                        mask=watershed_mask
                        )

        return res

    @staticmethod
    def relabel_image(image: np.ndarray) -> np.ndarray:
        """given a label image where some objects have been removed, relabel it with sequential integers

        Parameters
        ----------
        image : np.ndarray[np.uint32]
            image whose values identify which object each pixel corresponds to. the values may
            not be sequential integers.

        Returns
        -------
        image : np.ndarray[np.uint32]
            same as input, but the values are re-labled as sequential integers
        num_labels : int
            number of unique objects
        """
        output = np.empty_like(image)
        for i, v in enumerate(np.unique(image)):
            output[np.where(image == v)] = i
        return output, i

    def show(self, figsize=(10, 10)):
        import matplotlib.pyplot as plt
        plt.figure(figsize=figsize)

        plt.subplot(321)
        image(self.nuclei, ax=plt.gca(), size=20, bar=True)
        plt.title('Nuclei')

        plt.subplot(322)
        image(self.stain, ax=plt.gca(), size=20, bar=True)
        plt.title('Stain')

        plt.subplot(323)
        image(self.nuclei_thresholded, bar=False, ax=plt.gca())
        plt.title('Nuclei Thresholded')

        plt.subplot(324)
        image(self.mask, bar=False, ax=plt.gca())
        plt.title('Watershed Mask')

        plt.subplot(325)
        image(self.markers, size=20, cmap=plt.cm.nipy_spectral, ax=plt.gca())
        plt.title('Found: {} cells'.format(self.num_cells))

        plt.subplot(326)
        image(self.segmented, size=20, cmap=plt.cm.nipy_spectral, ax=plt.gca())
        plt.title('Segmented Cells')

        return plt.gca()



import numpy as np
import xarray as xr

from starfish.imagestack.imagestack import ImageStack
from starfish.imagestack.physical_coordinate_calculator import get_physical_coordinates_of_spot
from starfish.intensity_table.intensity_table import IntensityTable
from starfish.types import Coordinates, Features, Indices


def transfer_physical_coords_from_imagestack_to_intensity_table(image_stack: ImageStack,
                                                                intensity_table: IntensityTable
                                                                ) -> IntensityTable:
    """
    Transfers physical coordinates from an Imagestack's coordinates xarray to an intensity table
        1.) Creates three new coords on the intensity table (xc, yc, zc)
        2.) For every spot:
                Get pixel x,y values
                Calculate the physical x,y values
                Assign those values to the coords arrays for this spot
    """
    # Add three new coords to xarray (xc, yc, zc)
    num_features = intensity_table.sizes[Features.AXIS]
    intensity_table[Coordinates.X.value] = xr.DataArray(np.zeros(num_features, np.float32),
                                                        dims='features')
    intensity_table[Coordinates.Y.value] = xr.DataArray(np.zeros(num_features, np.float32),
                                                        dims='features')
    intensity_table[Coordinates.Z.value] = xr.DataArray(np.zeros(num_features, np.float32),
                                                        dims='features')
    for ind, spot in intensity_table.groupby(Features.AXIS):
        # Iterate through r, ch per spot
        for ch, _round in np.ndindex(spot.data.shape):
            # if non zero value set coords
            if spot[ch][_round].data == 0:
                continue
            # get pixel coords of this tile
            pixel_x = spot.coords[Indices.X].data
            pixel_y = spot.coords[Indices.Y].data
            pixel_z = spot.coords[Indices.Z].data
            tile_indices = {
                Indices.ROUND.value: _round,
                Indices.CH.value: ch,
                Indices.Z.value: pixel_z,
            }
            # Get corresponding physical coordinates
            physical_coords = get_physical_coordinates_of_spot(
                image_stack._coordinates,
                tile_indices,
                pixel_x,
                pixel_y,
                image_stack._tile_shape)
            # Assign to coordinates arrays
            intensity_table[Coordinates.X.value][ind] = physical_coords[0]
            intensity_table[Coordinates.Y.value][ind] = physical_coords[1]
            intensity_table[Coordinates.Z.value][ind] = physical_coords[2]
            break
    return intensity_table



import os
from typing import Type

import click
import numpy as np
from skimage.io import imread

from starfish.intensity_table.intensity_table import IntensityTable
from starfish.pipeline import AlgorithmBase, PipelineComponent
from . import label
from ._base import TargetAssignmentAlgorithm


class TargetAssignment(PipelineComponent):

    @classmethod
    def _get_algorithm_base_class(cls) -> Type[AlgorithmBase]:
        return TargetAssignmentAlgorithm

    @classmethod
    def _cli_run(cls, ctx, instance):
        output = ctx.obj["output"]
        intensity_table = ctx.obj["intensity_table"]
        label_image = ctx.obj["label_image"]
        assigned = instance.run(label_image, intensity_table)
        print(f"Writing intensities, including cell ids to {output}")
        assigned.save(os.path.join(output))

    @staticmethod
    @click.group("target_assignment")
    @click.option("--label-image", required=True, type=click.Path(exists=True))
    @click.option("--intensities", required=True, type=click.Path(exists=True))
    @click.option("-o", "--output", required=True)
    @click.pass_context
    def _cli(ctx, label_image, intensities, output):

        print('Assigning targets to cells...')
        ctx.obj = dict(
            component=TargetAssignment,
            output=output,
            intensity_table=IntensityTable.load(intensities),
            label_image=imread(label_image)
        )


TargetAssignment._cli_register()



import functools
import warnings
from collections import OrderedDict
from multiprocessing import Array as mp_array  # type: ignore
from typing import Callable, Sequence, Tuple

import numpy as np
import xarray as xr
from xarray import Variable

from starfish.types import Number


class MPDataArray:
    """Wrapper class for xarray.  It provides limited delegation to simplify ImageStack, but code
    external to ImageStack should transact in the enclosed xarray, as operations that involve
    special method names (e.g., __eq__) do not delegate correctly.

    This is necessary for us to stack an xarray on top of a numpy array that is stacked on top of a
    multiprocessing.Array object.  When we want to pass the xarray to worker processes using
    Python's multiprocessing module, we need to pass the underlying multiprocessing.Array object
    instead of the xarray or the numpy array.  However, there is no way to extract the
    multiprocessing.Array object back out of the numpy array or the xarray.  Therefore, we need to
    explicitly maintain a reference to it and keep the two items together.
    """
    def __init__(self, data: xr.DataArray, backing_mp_array: mp_array) -> None:
        self._data = data
        self._backing_mp_array = backing_mp_array

    @classmethod
    def from_shape_and_dtype(
            cls, shape: Sequence[int], dtype, initial_value: Number=None, *args, **kwargs
    ) -> "MPDataArray":
        np_array, backing_mp_array = np_array_backed_by_mp_array(shape, dtype)
        if initial_value is not None:
            np_array.fill(initial_value)
        xarray = xr.DataArray(np_array, *args, **kwargs)
        xarray.copy = functools.partial(replacement_copy, xarray.copy)
        return MPDataArray(xarray, backing_mp_array)

    @property
    def data(self) -> xr.DataArray:
        return self._data

    def __getattr__(self, item):
        return getattr(self._data, item)

    def __getitem__(self, key):
        return self._data[key]

    def __setitem__(self, key, value):
        self._data[key] = value

    def __deepcopy__(self, memodict={}):
        xarray_copy, backing_mp_array_copy = xr_deepcopy(self._data)
        return MPDataArray(xarray_copy, backing_mp_array_copy)


def np_array_backed_by_mp_array(
        shape: Sequence[int], dtype) -> Tuple[np.ndarray, mp_array]:
    """Returns a np_array backed by a multiproceessing.Array buffer."""
    ctype_type = np.ctypeslib.as_ctypes(np.empty((1,), dtype=np.dtype(dtype))).__class__
    length = int(np.product(shape))  # the cast to int is required by multiprocessing.Array.
    backing_array = mp_array(ctype_type, length)
    unshaped_np_array = np.frombuffer(backing_array.get_obj(), dtype)
    shaped_np_array = unshaped_np_array.reshape(shape)

    return shaped_np_array, backing_array


def xr_deepcopy(source: xr.DataArray) -> Tuple[xr.DataArray, mp_array]:
    """Replacement for xr.DataArray's deepcopy method.  Returns a deep copy of the input xarray
    backed by a multiprocessing.Array buffer.
    """
    shaped_np_array, backing_array = np_array_backed_by_mp_array(
        source.variable.shape, source.variable.dtype)

    shaped_np_array[:] = source.variable.data

    variable = Variable(
        # dims is an immutable tuple, so it doesn't need to be deep-copyied.  see implementation of
        # Variable.__deepcopy__ for context.
        source.variable.dims,
        shaped_np_array,
        # attrs and encoding are deep copied in the constructor.
        source.variable.attrs,
        source.variable.encoding,
    )
    coords = OrderedDict((k, v.copy(deep=True))
                         for k, v in source._coords.items())

    result = xr.DataArray(
        variable,
        coords=coords,
        name=source.name,
        fastpath=True,
    )

    return result, backing_array


def replacement_copy(orig_copy: Callable, deep=True):
    if deep:
        warnings.warn(
            "Calling deepcopy with ImageStack's xarrays results in arrays that cannot be shared "
            "across processes.  Use mp_dataarray.xr_deepcopy if you need that behavior."
        )
    return orig_copy(deep)



import pytest
from pkg_resources import resource_filename

from starfish.codebook._format import DocumentKeys
from .util import SpaceTxValidator

package_name = "sptx_format"
codebook_schema_path = resource_filename(package_name, "schema/codebook/codebook.json")
validator = SpaceTxValidator(codebook_schema_path)


def test_codebook():
    codebook_example_path = resource_filename(
        package_name, "examples/codebook/codebook.json")
    assert validator.validate_file(codebook_example_path)


def test_diagonal_codebook():
    codebook_example_path = resource_filename(
        package_name, "examples/codebook/codebook_diagonal.json")
    assert validator.validate_file(codebook_example_path)


def test_diagonal_codebook_full_values():
    codebook_example_path = resource_filename(
        package_name, "examples/codebook/codebook_diagonal_inferred_value.json")
    assert validator.validate_file(codebook_example_path)


def test_codebook_missing_channel_raises_validation_error():
    codebook_example_path = resource_filename(
        package_name, "examples/codebook/codebook.json")
    codebook = validator.load_json(codebook_example_path)
    del codebook[DocumentKeys.MAPPINGS_KEY][0]['codeword'][0]['c']
    with pytest.warns(UserWarning):
        assert not validator.validate_object(codebook)



import collections
from typing import Mapping

from starfish.types import Indices

_DimensionMetadata = collections.namedtuple("_DimensionMetadata", ['order', 'required'])

AXES_DATA: Mapping[Indices, _DimensionMetadata] = {
    Indices.ROUND: _DimensionMetadata(0, True),
    Indices.CH: _DimensionMetadata(1, True),
    Indices.Z: _DimensionMetadata(2, False),
}
N_AXES = max(data.order for data in AXES_DATA.values()) + 1



import numpy as np

from starfish.spots._detector.blob import BlobDetector
from starfish.types import Features, Indices
from starfish.util.synthesize import SyntheticData


def test_round_trip_synthetic_data():
    np.random.seed(0)

    sd = SyntheticData(
        n_ch=2,
        n_round=3,
        n_spots=1,
        n_codes=4,
        n_photons_background=0,
        background_electrons=0,
        camera_detection_efficiency=1.0,
        gray_level=1,
        ad_conversion_bits=16,
        point_spread_function=(2, 2, 2),
    )

    codebook = sd.codebook()
    intensities = sd.intensities(codebook=codebook)
    spots = sd.spots(intensities=intensities)
    spots_mp = spots.max_proj(Indices.CH, Indices.ROUND)
    spots_mp_numpy = spots_mp._squeezed_numpy(Indices.CH, Indices.ROUND)
    gsd = BlobDetector(min_sigma=1, max_sigma=4, num_sigma=5, threshold=0)
    calculated_intensities = gsd.run(spots, blobs_image=spots_mp_numpy)
    decoded_intensities = codebook.metric_decode(
        calculated_intensities,
        max_distance=1,
        min_intensity=0,
        norm_order=2
    )

    # applying the gaussian blur to the intensities causes them to be reduced in magnitude, so
    # they won't be the same size, but they should be in the same place, and decode the same
    # way
    spot1, ch1, round1 = np.where(intensities.values)
    spot2, ch2, round2 = np.where(calculated_intensities.values)
    assert np.array_equal(spot1, spot2)
    assert np.array_equal(ch1, ch2)
    assert np.array_equal(round1, round2)
    assert np.array_equal(
        intensities.coords[Features.TARGET],
        decoded_intensities.coords[Features.TARGET]
    )


# @pytest.mark.skip('long-running integration test, for debugging only')
def test_medium_synthetic_stack():
    np.random.seed(0)

    n_z = 40
    height = 300
    width = 400
    sigma = 2

    sd = SyntheticData(
        n_round=4,
        n_ch=4,
        n_z=n_z,
        height=height,
        width=width,
        n_spots=100,
        n_codes=10,
        point_spread_function=(sigma, sigma, sigma),
    )

    codebook = sd.codebook()
    intensities = sd.intensities(codebook=codebook)

    # some spots won't be detected properly because they will spread outside the image when blurred,
    # so we'll remove those from intensities before we generate spots.

    spot_radius = sigma * np.sqrt(2)  # this is the radius of the spot in pixels

    valid_z = np.logical_and(intensities.z.values > spot_radius,
                             intensities.z.values < (n_z - spot_radius))
    valid_y = np.logical_and(intensities.y.values > spot_radius,
                             intensities.y.values < (height - spot_radius))
    valid_x = np.logical_and(intensities.x.values > spot_radius,
                             intensities.x.values < (width - spot_radius))

    valid_locations = valid_z & valid_y & valid_x
    intensities = intensities[np.where(valid_locations)]
    spots = sd.spots(intensities=intensities)
    spots_mp = spots.max_proj(Indices.CH, Indices.ROUND)
    spots_mp_numpy = spots_mp._squeezed_numpy(Indices.CH, Indices.ROUND)
    gsd = BlobDetector(min_sigma=1, max_sigma=4, num_sigma=5, threshold=1e-4)
    calculated_intensities = gsd.run(spots, blobs_image=spots_mp_numpy)
    calculated_intensities = codebook.metric_decode(
        calculated_intensities, max_distance=1, min_intensity=0, norm_order=2
    )

    # spots are detected in a different order that they're generated; sorting makes comparison easy
    sorted_intensities = intensities.sortby([Indices.Z.value, Indices.Y.value, Indices.X.value])
    sorted_calculated_intensities = calculated_intensities.sortby(
        [Indices.Z.value, Indices.Y.value, Indices.X.value]
    )

    # verify that the spots are all detected, and decode to the correct targets
    assert np.array_equal(
        sorted_intensities[Features.TARGET].values,
        sorted_calculated_intensities[Features.TARGET].values
    )



from typing import List, Optional, Tuple, Union

import click
import numpy as np
import pandas as pd
import xarray as xr
from scipy.ndimage import label
from skimage.feature import peak_local_max
from skimage.measure import regionprops
from sympy import Line, Point
from tqdm import tqdm

from starfish.imagestack.imagestack import ImageStack
from starfish.intensity_table.intensity_table import IntensityTable
from starfish.types import Features, Indices, Number, SpotAttributes
from ._base import SpotFinderAlgorithmBase
from .detect import detect_spots


# TODO ambrosejcarr, ttung: https://github.com/spacetx/starfish/issues/708
# Currently, spot finders cannot propagate state, which makes the flow for this
# spot finder confusing. One would expect to have access to the private parameters
# however, they are lost due to the memory-space forking induced by multi-processing.

class LocalMaxPeakFinder(SpotFinderAlgorithmBase):
    def __init__(
        self, min_distance: int, stringency: int, min_obj_area: int, max_obj_area: int,
        threshold: Optional[Number]=None, measurement_type: str='max',
        min_num_spots_detected: int=3, is_volume: bool=False, verbose: bool=True
    ) -> None:
        """2-dimensional local-max peak finder that wraps skimage.feature.peak_local_max

        Parameters:
        min_distance : int
            Minimum number of pixels separating peaks in a region of 2 * min_distance + 1
            (i.e. peaks are separated by at least min_distance). To find the maximum number of
            peaks, use min_distance=1.
        stringency : int
        min_obj_area : int
        max_obj_area : int
        threshold : Optional[Number]
        measurement_type : str, {'max', 'mean'}
            default 'max' calculates the maximum intensity inside the object
        min_num_spots_detected : int
            When fewer than this number of spots are detected, spot searching for higher threshold
            values. (default = 3)
        is_volume : bool
            Not supported. For 3d peak detection please use TrackpyLocalMaxPeakFinder.
            (default=False)
        verbose : bool
            If True, report the percentage completed (default = False) during processing

        See Also
        --------
        http://scikit-image.org/docs/dev/api/skimage.feature.html#skimage.feature.peak_local_max
        """

        self.min_distance = min_distance
        self.stringency = stringency
        self.min_obj_area = min_obj_area
        self.max_obj_area = max_obj_area
        self.threshold = threshold
        self.min_num_spots_detected = min_num_spots_detected

        self.measurement_function = self._get_measurement_function(measurement_type)

        self.is_volume = is_volume
        if self.is_volume:
            raise ValueError('LocalMaxPeakFinder only works for 2D data, for 3D data, '
                             'please use TrackpyLocalMaxPeakFinder')

        self.verbose = verbose

        # these parameters are useful for debugging spot-calls
        self._thresholds: Optional[np.ndarray] = None
        self._spot_counts: Optional[List[int]] = None
        self._grad = None
        self._spot_props = None
        self._labels = None

    def _compute_num_spots_per_threshold(self, img: np.ndarray) -> Tuple[np.ndarray, List[int]]:
        """Computes the number of detected spots for each threshold

        Parameters
        ----------
        img : np.ndarray
            The image in which to count spots

        Returns
        -------
        np.ndarray :
            thresholds
        List[int] :
            spot counts
        """

        # thresholds to search over
        thresholds = np.linspace(img.min(), img.max(), num=100)

        # number of spots detected at each threshold
        spot_counts = []

        # where we stop our threshold search
        stop_threshold = None

        if self.verbose:
            threshold_iter = tqdm(thresholds)
            print('Determining optimal threshold ...')
        else:
            threshold_iter = thresholds

        for stop_index, threshold in enumerate(threshold_iter):
            spots = peak_local_max(
                img,
                min_distance=self.min_distance,
                threshold_abs=threshold,
                exclude_border=False,
                indices=True,
                num_peaks=np.inf,
                footprint=None,
                labels=None
            )

            # stop spot finding when the number of detected spots falls below min_num_spots_detected
            if len(spots) <= self.min_num_spots_detected:
                stop_threshold = threshold
                if self.verbose:
                    print(f'Stopping early at threshold={threshold}. Number of spots fell below: '
                          f'{self.min_num_spots_detected}')
                break
            else:
                spot_counts.append(len(spots))

        if stop_threshold is None:
            stop_threshold = thresholds.max()

        if len(thresholds > 1):
            thresholds = thresholds[:stop_index]
            spot_counts = spot_counts[:stop_index]

        # TODO ambrosejcarr: these two lists should be a dict
        self._thresholds = thresholds
        self._spot_counts = spot_counts

        return thresholds, spot_counts

    def _select_optimal_threshold(self, thresholds: np.ndarray, spot_counts: List[int]) -> float:

        # calculate the gradient of the number of spots
        grad = np.gradient(spot_counts)
        self._grad = grad
        optimal_threshold_index = np.argmin(grad)

        # only consider thresholds > than optimal threshold
        thresholds = thresholds[optimal_threshold_index:]
        grad = grad[optimal_threshold_index:]

        # if all else fails, return 0.
        selected_thr = 0

        if len(thresholds) > 1:

            distances = []

            # create a line whose end points are the threshold and and corresponding gradient value
            # for spot_counts corresponding to the threshold
            start_point = Point(thresholds[0], grad[0])
            end_point = Point(thresholds[-1], grad[-1])
            line = Line(start_point, end_point)

            # calculate the distance between all points and the line
            for k in range(len(thresholds)):
                p = Point(thresholds[k], grad[k])
                dst = line.distance(p)
                distances.append(dst.evalf())

            # remove the end points
            thresholds = thresholds[1:-1]
            distances = distances[1:-1]

            # select the threshold that has the maximum distance from the line
            # if stringency is passed, select a threshold that is n steps higher, where n is the
            # value of stringency
            if distances:
                thr_idx = np.argmax(np.array(distances))

                if thr_idx + self.stringency < len(thresholds):
                    selected_thr = thresholds[thr_idx + self.stringency]
                else:
                    selected_thr = thresholds[thr_idx]

        return selected_thr

    def _compute_threshold(self, img: Union[np.ndarray, xr.DataArray]) -> float:
        """Finds spots on a number of thresholds then selects and returns the optimal threshold

        Parameters
        ----------
        img: Union[np.ndarray, xr.DataArray]
            data array in which spots should be detected and over which to compute different
            intensity thresholds

        Returns
        -------
        Number :  #TODO ambrosejcarr this should probably be a float
            The intensity threshold
        """
        thresholds, spot_counts = self._compute_num_spots_per_threshold(img)
        threshold = self._select_optimal_threshold(thresholds, spot_counts)
        return threshold

    def image_to_spots(self, data_image: Union[np.ndarray, xr.DataArray]) -> SpotAttributes:
        """measure attributes of spots detected by binarizing the image using the selected threshold

        Parameters
        ----------
        data_image: Union[np.ndarray, xr.DataArray]
            image from which spots should be extracted

        Returns
        -------
        SpotAttributes
            Attributes for each detected spot
        """

        if self.threshold is None:
            self.threshold = self._compute_threshold(data_image)

        # identify each spot's size by binarizing and calculating regionprops
        masked_image = data_image[:, :] > self.threshold
        labels = label(masked_image)[0]
        spot_props = regionprops(labels)

        # mask spots whose areas are too small or too large
        for spot_prop in spot_props:
            if spot_prop.area < self.min_obj_area or spot_prop.area > self.max_obj_area:
                masked_image[spot_prop.coords[:, 0], spot_prop.coords[:, 1]] = 0

        # store re-calculated regionprops and labels based on the area-masked image
        self._labels = label(masked_image)[0]
        self._spot_props = regionprops(labels)

        if self.verbose:
            print('computing final spots ...')

        self._spot_coords = peak_local_max(
            data_image,
            min_distance=self.min_distance,
            threshold_abs=self.threshold,
            exclude_border=False,
            indices=True,
            num_peaks=np.inf,
            footprint=None,
            labels=self._labels
        )

        # TODO how to get the radius? unlikely that this can be pulled out of
        # self._spot_props, since the last call to peak_local_max can find multiple
        # peaks per label
        res = {Indices.X.value: self._spot_coords[:, 1],
               Indices.Y.value: self._spot_coords[:, 0],
               Indices.Z.value: np.zeros(len(self._spot_coords)),
               Features.SPOT_RADIUS: 1,
               Features.SPOT_ID: np.arange(self._spot_coords.shape[0]),
               Features.INTENSITY: data_image[self._spot_coords[:, 0], self._spot_coords[:, 1]]
               }

        return SpotAttributes(pd.DataFrame(res))

    def run(
            self,
            data_stack: ImageStack,
            blobs_image: Optional[Union[np.ndarray, xr.DataArray]] = None,
            reference_image_from_max_projection: bool = False,
    ) -> IntensityTable:
        """Find spots.

        Parameters
        ----------
        data_stack : ImageStack
            Stack where we find the spots in.
        blobs_image : Union[np.ndarray, xr.DataArray]
            If provided, spots will be found in the blobs image, and intensities will be measured
            across hybs and channels. Otherwise, spots are measured independently for each channel
            and round.
        reference_image_from_max_projection : bool
            if True, compute a reference image from the maximum projection of the channels and
            z-planes

        Returns
        -------
        IntensityTable :
            IntensityTable containing decoded spots
        """
        intensity_table = detect_spots(
            data_stack=data_stack,
            spot_finding_method=self.image_to_spots,
            reference_image=blobs_image,
            reference_image_from_max_projection=reference_image_from_max_projection,
            measurement_function=self.measurement_function,
            radius_is_gyration=False,
            is_volume=self.is_volume
        )

        return intensity_table

    @staticmethod
    @click.command("LocalMaxPeakFinder")
    @click.option(
        "--min-distance", default=4, type=int,
        help="Minimum spot size (in number of pixels deviation)")
    @click.option(
        "--min-obj-area", default=6, type=int,
        help="Maximum spot size (in number of pixels")
    @click.option(
        "--max_obj_area", default=300, type=int,
        help="Maximum spot size (in number of pixels)")
    @click.option(
        "--stringency", default=0, type=int,
        help="Number of indices in threshold list to look past "
             "for the threhsold finding algorithm")
    @click.option(
        "--threshold", default=None, type=float,
        help="Threshold on which to threshold "
             "image prior to spot detection")
    @click.option(
        "--min-num-spots-detected", default=3, type=int,
        help="Minimum number of spots detected at which to stop a"
             "utomatic thresholding algorithm")
    @click.option(
        "--measurement-type", default='max', type=str,
        help="How to aggregate pixel intensities in a spot")
    @click.option(
        "--is-volume", default=False, action='store_false', help="Find spots in 3D or not")
    @click.option(
        "--verbose", default=True, action='store_true', help="Verbosity flag")
    @click.pass_context
    def _cli(ctx, min_distance, min_obj_area, max_obj_area, stringency, threshold,
             min_num_spots_detected, measurement_type, is_volume, verbose):
        instance = LocalMaxPeakFinder(min_distance, min_obj_area, max_obj_area,
                                      stringency, threshold,
                                      min_num_spots_detected, measurement_type, is_volume, verbose)
        ctx.obj["component"]._cli_run(ctx, instance)



"""
Tests for IntensityTable.empty_intensity_table method
"""

import numpy as np
import pandas as pd
import pytest

from starfish import IntensityTable
from starfish.types import Features, Indices, SpotAttributes


def test_intensity_table_can_be_created_from_spot_attributes():
    """
    This test creates an IntensityTable from spot attributes, and verifies that the size matches
    what was requested and that the values are all zero.
    """

    # input has two spots
    spot_attributes = SpotAttributes(
        pd.DataFrame(
            data=np.array(
                [[1, 1, 1, 1],
                 [2, 2, 2, 1]]
            ),
            columns=[Indices.Z, Indices.Y, Indices.X, Features.SPOT_RADIUS]
        )
    )

    intensities = IntensityTable.empty_intensity_table(
        spot_attributes,
        n_ch=1,
        n_round=3
    )

    assert intensities.sizes[Indices.CH] == 1
    assert intensities.sizes[Indices.ROUND] == 3
    assert intensities.sizes[Features.AXIS] == 2
    assert np.all(intensities.values == 0)


def test_from_spot_attributes_throws_type_error_when_passed_a_dataframe():
    """SpotAttributes should be passed instead."""
    # input has two spots
    not_spot_attributes = pd.DataFrame(
        data=np.array(
            [[1, 1, 1, 1],
             [2, 2, 2, 1]]
        ),
        columns=[Indices.Z, Indices.Y, Indices.X, Features.SPOT_RADIUS]
    )

    with pytest.raises(TypeError):
        IntensityTable.empty_intensity_table(not_spot_attributes, n_ch=1, n_round=3)



import warnings
from typing import Mapping, Tuple, Union

import numpy as np
import pytest
from skimage import img_as_float32
from slicedimage import ImageFormat

from starfish.errors import DataFormatWarning
from starfish.experiment.builder import FetchedTile, TileFetcher
from starfish.imagestack.imagestack import ImageStack
from starfish.types import Coordinates, Number

NUM_ROUND = 2
NUM_CH = 2
NUM_Z = 2
HEIGHT = 10
WIDTH = 10


class OnesTilesByDtype(FetchedTile):
    def __init__(self, dtype: np.number) -> None:
        super().__init__()
        self._dtype = dtype

    @property
    def shape(self) -> Tuple[int, ...]:
        return HEIGHT, WIDTH

    @property
    def coordinates(self) -> Mapping[Union[str, Coordinates], Union[Number, Tuple[Number, Number]]]:
        return {
            Coordinates.X: (0.0, 0.0001),
            Coordinates.Y: (0.0, 0.0001),
            Coordinates.Z: (0.0, 0.0001),
        }

    @property
    def format(self) -> ImageFormat:
        return ImageFormat.TIFF

    def tile_data(self) -> np.ndarray:
        return np.ones((HEIGHT, WIDTH), dtype=self._dtype)


class CornerDifferentDtype(TileFetcher):
    """
    TileFetcher that fetches tiles all of the same type, except the round=0,ch=0,z=0 corner, which
    is a different type.  All the tiles are initialized with ones.

    Parameters
    ----------
    dtype : np.number
        The data type of all the tiles except the hyd=0,ch=0,z=0 corner.
    corner_dtype
        The data type of the tile in the hyd=0,ch=0,z=0 corner.
    """
    def __init__(self, dtype: np.number, corner_dtype: np.number) -> None:
        self._dtype = dtype
        self._corner_dtype = corner_dtype

    def get_tile(self, fov: int, round_: int, ch: int, z: int) -> FetchedTile:
        if round_ == 0 and ch == 0 and z == 0:
            dtype = self._corner_dtype
        else:
            dtype = self._dtype

        return OnesTilesByDtype(dtype)


def test_multiple_tiles_of_different_kind():
    with pytest.raises(TypeError):
        ImageStack.synthetic_stack(
            NUM_ROUND, NUM_CH, NUM_Z,
            HEIGHT, WIDTH,
            tile_fetcher=CornerDifferentDtype(np.uint32, np.float32),
        )


def test_multiple_tiles_of_same_dtype():
    stack = ImageStack.synthetic_stack(
        NUM_ROUND, NUM_CH, NUM_Z,
        HEIGHT, WIDTH,
        tile_fetcher=CornerDifferentDtype(np.uint32, np.uint32),
    )
    expected = np.ones(
        (NUM_ROUND,
         NUM_CH,
         NUM_Z,
         HEIGHT,
         WIDTH), dtype=np.uint32)
    assert np.array_equal(stack.xarray, img_as_float32(expected))


def test_int_type_promotion():
    with warnings.catch_warnings(record=True) as warnings_:
        stack = ImageStack.synthetic_stack(
            NUM_ROUND, NUM_CH, NUM_Z,
            HEIGHT, WIDTH,
            tile_fetcher=CornerDifferentDtype(np.int32, np.int8),
        )
        assert len(warnings_) == 2
        assert issubclass(warnings_[0].category, DataFormatWarning)
        assert issubclass(warnings_[1].category, UserWarning)
    expected = img_as_float32(np.ones(
        (NUM_ROUND,
         NUM_CH,
         NUM_Z,
         HEIGHT,
         WIDTH), dtype=np.int32))
    corner = img_as_float32(np.ones(
        (HEIGHT,
         WIDTH), dtype=np.int8))
    expected[0, 0, 0] = corner
    assert np.array_equal(stack.xarray, img_as_float32(expected))


def test_uint_type_promotion():
    with warnings.catch_warnings(record=True) as warnings_:
        stack = ImageStack.synthetic_stack(
            NUM_ROUND, NUM_CH, NUM_Z,
            HEIGHT, WIDTH,
            tile_fetcher=CornerDifferentDtype(np.uint32, np.uint8),
        )
        assert len(warnings_) == 2
        assert issubclass(warnings_[0].category, DataFormatWarning)
        assert issubclass(warnings_[1].category, UserWarning)
    expected = img_as_float32(np.ones(
        (NUM_ROUND,
         NUM_CH,
         NUM_Z,
         HEIGHT,
         WIDTH), dtype=np.uint32))
    corner = img_as_float32(np.ones(
        (HEIGHT,
         WIDTH), dtype=np.uint8))
    expected[0, 0, 0] = corner
    assert np.array_equal(stack.xarray, img_as_float32(expected))


def test_float_type_demotion():
    with warnings.catch_warnings(record=True) as warnings_:
        stack = ImageStack.synthetic_stack(
            NUM_ROUND, NUM_CH, NUM_Z,
            HEIGHT, WIDTH,
            tile_fetcher=CornerDifferentDtype(np.float64, np.float32),
        )
        assert len(warnings_) == 2
        assert issubclass(warnings_[0].category, DataFormatWarning)
        assert issubclass(warnings_[1].category, UserWarning)
    expected = np.ones(
        (NUM_ROUND,
         NUM_CH,
         NUM_Z,
         HEIGHT,
         WIDTH), dtype=np.float64)
    assert np.array_equal(stack.xarray, expected)



from typing import Callable, Container, Optional


class StripArguments:
    """
    Class to strip out arguments to a function call.  This is spiritually the opposite of
    functools.partial.  This proxy will remove some arguments and call the nested callable.

    This is used in scenarios where factory methods and lambdas cannot be used, such as dispatching
    work with multiprocessing.  Inputs to multiprocessing must be pickle-able, and factory methods
    and lambdas always generate a lexical scope (i.e., locals()) that cannot be picked.

    This class is initialized with a callable, positional argument positions to remove, and keyword
    arguments to remove.

    For instance, if someone expects a callable that accepts as its parameters (str, int, obj), but
    the callable you have only accepts an int (and does not need the str, obj arguments), then wrap
    the object with _StripArguments(my_callable, positional_arguments_removed=(0, 2)).
    """
    def __init__(
            self,
            func: Callable,
            *,
            positional_arguments_removed: Optional[Container[int]]=None,
            keyword_arguments_removed: Container[set]=None,
    ) -> None:
        self.func = func
        self.positional_arguments_removed = positional_arguments_removed or list()
        self.keyword_arguments_removed = keyword_arguments_removed or set()

    def __call__(self, *args, **kwargs):
        called_args = [
            arg
            for ix, arg in enumerate(args)
            if ix not in self.positional_arguments_removed
        ]
        for keyword_argument_removed in self.keyword_arguments_removed:
            kwargs.remove(keyword_argument_removed)
        return self.func(*called_args, **kwargs)



from typing import Set, Tuple, Union

import numpy as np
import xarray as xr
from skimage.morphology import binary_opening, disk

from starfish.types import Indices, Number


def bin_thresh(img: np.ndarray, thresh: Number) -> np.ndarray:
    """
    Performs binary thresholding of an image

    Parameters
    ----------
    img : np.ndarray
        Image to filter.
    thresh : int
        Pixel values >= thresh are set to 1, else 0.

    Returns
    -------
    np.ndarray :
        Binarized image, same shape as input

    """
    res = img >= thresh
    return res


def bin_open(img: np.ndarray, disk_size: int) -> np.ndarray:
    """
    Performs binary opening of an image

    img : np.ndarray
        Image to filter.
    masking_radius : int
        Radius of the disk-shaped structuring element.

    Returns
    -------
    np.ndarray :
        Filtered image, same shape as input

    """
    selem = disk(disk_size)
    res = binary_opening(img, selem)
    return res


def gaussian_kernel(shape: Tuple[int, int]=(3, 3), sigma: float=0.5):
    """
    Returns a gaussian kernel of specified shape and standard deviation.
    This is a simple python implementation of Matlab's fspecial('gaussian',[shape],[sigma])

    Parameters
    ----------
    shape : Tuple[int] (default = (3, 3))
        Kernel shape.
    sigma : float (default = 0.5)
        Standard deviation of gaussian kernel.

    Parameters
    ----------
    np.ndarray :
        Gaussian kernel.
    """
    m, n = [int((ss - 1.) / 2.) for ss in shape]
    y, x = np.ogrid[-m:m + 1, -n:n + 1]
    kernel = np.exp(-(x * x + y * y) / (2. * sigma * sigma))
    kernel[kernel < np.finfo(kernel.dtype).eps * kernel.max()] = 0
    sumh = kernel.sum()
    if sumh != 0:
        kernel /= sumh
    return kernel


def validate_and_broadcast_kernel_size(
        sigma: Union[Number, Tuple[Number, ...]],
        is_volume: bool
) -> Tuple[Number, ...]:
    """
    Check that the provided sigma is of the right dimensionality, and if necessary, broadcast it
    to full dimensionality.

    Parameters
    ----------
    sigma : Union[Number, Tuple[Number]]
    is_volume : bool

    Returns
    -------
    Tuple[Number] :
        2-d or 3-d kernel size.

    """
    if isinstance(sigma, tuple):
        message = ("if passing an anisotropic kernel, the dimensionality must match the data "
                   "shape ({shape}), not {passed_shape}")
        if is_volume and len(sigma) != 3:
            raise ValueError(message.format(shape=3, passed_shape=len(sigma)))
        if not is_volume and len(sigma) != 2:
            raise ValueError(message.format(shape=2, passed_shape=len(sigma)))
        valid_sigma = sigma
    else:
        if is_volume:
            valid_sigma = (sigma,) * 3
        else:
            valid_sigma = (sigma,) * 2

    return valid_sigma


def preserve_float_range(
        array: Union[xr.DataArray, np.ndarray],
        rescale: bool=False) -> Union[xr.DataArray, np.ndarray]:
    """
    Clips values below zero to zero. If values above one are detected, clips them
    to 1 unless `rescale` is True, in which case the input is scaled by
    the max value and the dynamic range is preserved.

    Parameters
    ----------
    array : Union[xr.DataArray, np.ndarray]
        Array whose values should be in the interval [0, 1] but may not be.
    rescale: bool
        If true, scale values by the max.

    Returns
    -------
    array : Union[xr.DataArray, np.ndarray]
        Array whose values are in the interval [0, 1].

    """
    array = array.copy()
    if isinstance(array, xr.DataArray):
        data = array.values
    else:
        data = array
    if np.any(data < 0):
        data[array < 0] = 0
    if np.any(array > 1):
        if rescale:
            data /= data.max()
        else:
            data[array > 1] = 1
    return array.astype(np.float32)


def determine_axes_to_group_by(is_volume: bool) -> Set[Indices]:
    """map is_volume to axes to group by when applying a function over an ImageStack"""
    if is_volume:
        return {Indices.ROUND, Indices.CH}
    else:
        return {Indices.ROUND, Indices.CH, Indices.Z}



from typing import Optional, Sequence

import matplotlib.axes
import matplotlib.pyplot as plt

from starfish.types import Number
from .util import annotate_axis


def _plot_threshold_indicator(x_location: Number, ax: matplotlib.axes.Axes) -> matplotlib.axes.Axes:
    """plots a vertical dotted line to represent a threshold applied to the histogram"""
    ymin, ymax = ax.get_ylim()
    ax.vlines(x=x_location, ymin=ymin, ymax=ymax, linestyles='--', colors='r')
    ax.set_ylim(ymin, ymax)  # reset the axis. Matplotlib expands when using vlines
    return ax


def histogram(
        data_vector: Sequence,
        ax: Optional[matplotlib.axes.Axes]=None,
        bins: int=25,
        threshold: Optional[Number]=None,
        title: Optional[str]='',
        xlabel: Optional[str]='',
        ylabel: Optional[str]='',
        log=True,
        **kwargs,
) -> matplotlib.axes.Axes:
    """Plot the distribution of spot areas

    Parameters
    ----------
    data_vector: Sequence
        data vector to be summarized
    ax : Optional[matplotlib.axes.Axes]
        Axis to plot in. If not provided, defaults to current axis.
    bins : int
        Number of bins. Default 25.
    log : bool
        Whether to plot the y-axis of the histogram in log scale.
    threshold: Optional[Number]
        If provided, plot a vertical line indicating where a threshold was selected for a given
        assay
    title, xlabel, ylabel : Optional[str]
        Labels to add to the title, x, and y labels of the plot
    kwargs : Dict
        additional keyword arguments to pass to matplotlib.pyplot.hist

    Returns
    -------
    matplotlib.axes.Axes :
        The axis containing the plot

    """
    ax = ax if ax is not None else plt.gca()
    ax.hist(data_vector, bins=bins, log=log, **kwargs)
    annotate_axis(
        ax,
        title=title,
        ylabel=ylabel,
        xlabel=xlabel,
    )
    if threshold is not None:
        _plot_threshold_indicator(threshold, ax)

    return ax



import functools
from typing import Callable, Optional, Set


def try_import(allowable_module_names: Optional[Set[str]]=None) -> Callable:
    """
    Decorator to apply to a method.  If one of the modules in `allowable_module_names` fail to
    import, raise a friendly error message.  If `allowable_module_names` is None, then all failed
    imports raise a friendly error message.

    Enables large and peripheral dependencies to be excluded from the build.
    """
    def _try_import_decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                func(*args, **kwargs)
            except ImportError as ex:
                module_name = ex.name

                if allowable_module_names is None or module_name in allowable_module_names:
                    raise ImportError(
                        f"{module_name} is an optional dependency of starfish. Please install "
                        f"{module_name} and its dependencies to use this functionality."
                    ) from ex
                else:
                    raise

        return wrapper

    return _try_import_decorator



"""
Tests for IntensityTable.save and IntensityTable.load methods
"""

import os
import tempfile

import numpy as np

from starfish import ImageStack, IntensityTable


def test_intensity_table_serialization():
    """
    Test that an IntensityTable can be saved to disk, and that when it is reloaded, the data is
    unchanged
    """

    # create an IntensityTable
    data = np.zeros(100, dtype=np.float32).reshape(1, 5, 2, 2, 5)
    image_stack = ImageStack.from_numpy_array(data)
    intensities = IntensityTable.from_image_stack(image_stack)

    # dump it to disk
    tempdir = tempfile.mkdtemp()
    filename = os.path.join(tempdir, 'test.nc')
    intensities.save(filename)

    # verify the data has not changed
    loaded = intensities.load(filename)
    assert intensities.equals(loaded)



from typing import Union

from ._constants import (
    Coordinates,
    Features,
    Indices,
    PHYSICAL_COORDINATE_DIMENSION,
    PhysicalCoordinateTypes,
)
from ._spot_attributes import SpotAttributes

Number = Union[int, float]



from typing import Tuple

import numpy as np

from starfish.imagestack.imagestack import ImageStack
from starfish.intensity_table.intensity_table import IntensityTable
from starfish.types import Indices


def test_reshaping_between_stack_and_intensities():
    """
    transform an pixels of an ImageStack into an IntensityTable and back again, then verify that
    the created Imagestack is the same as the original
    """
    np.random.seed(777)
    image = ImageStack.from_numpy_array(np.random.rand(1, 2, 3, 4, 5).astype(np.float32))
    pixel_intensities = IntensityTable.from_image_stack(image, 0, 0, 0)
    image_shape = (image.shape['z'], image.shape['y'], image.shape['x'])
    image_from_pixels = pixel_intensities_to_imagestack(pixel_intensities, image_shape)
    assert np.array_equal(image.xarray, image_from_pixels.xarray)


def pixel_intensities_to_imagestack(
        intensities: IntensityTable, image_shape: Tuple[int, int, int]
) -> ImageStack:
    """Re-create the pixel intensities from an IntensityTable

    Parameters
    ----------
    intensities : IntensityTable
        intensities to transform into an ImageStack
    image_shape : Tuple[int, int, int]
        the dimensions of z, y, and x for the original image that the intensity table was generated
        from

    Returns
    -------
    ImageStack :
        ImageStack containing Intensity information

    """
    # reverses the process used to produce the intensity table in to_pixel_intensities
    data = intensities.values.reshape([
        *image_shape,
        intensities.sizes[Indices.CH],
        intensities.sizes[Indices.ROUND]])
    data = data.transpose(4, 3, 0, 1, 2)
    return ImageStack.from_numpy_array(data)



"""
Tests for starfish.spots._detector.combine_adjacent_features.TargetsMap
"""

import numpy as np

from starfish.spots._detector.combine_adjacent_features import TargetsMap


def test_targets_map():
    """Test that TargetsMap provides an invertible map between string names and integer IDs"""

    targets = np.array(['None', 'test', 'test_2', 'I_Am_A-Gene', 'I.am.an.R.user'])
    target_map = TargetsMap(targets)

    encoded = target_map.targets_as_int(targets)
    decoded = target_map.targets_as_str(encoded)

    assert np.array_equal(decoded, targets)



"""
Tests for CombineAdjacentFeatures._intensities_to_decoded_image method
"""

from typing import Tuple

import numpy as np

from starfish import ImageStack, IntensityTable
from starfish.spots._detector.combine_adjacent_features import CombineAdjacentFeatures, TargetsMap
from starfish.types import Features


def decoded_intensity_table_factory() -> Tuple[IntensityTable, np.ndarray]:
    """
    Create an IntensityTable that has gene labels, including null labels. The data doesn't matter,
    so will use np.zeros
    """
    data = np.zeros((1, 1, 2, 3, 3), dtype=np.float32)
    labels = np.array(
        [[[0, 1, 1],
          [0, 2, 2],
          [1, 1, 1]],
         [[0, 1, 1],
          [1, 1, 1],
          [0, 1, 2]]],
        dtype='<U3'
    )
    labels_with_nan = labels.copy()
    labels_with_nan[labels == '0'] = 'nan'

    # create an intensity table and add the labels
    image_stack = ImageStack.from_numpy_array(data)
    intensities = IntensityTable.from_image_stack(image_stack)
    intensities[Features.TARGET] = (Features.AXIS, np.ravel(labels_with_nan))

    # label the third column of this data as failing filters
    passes_filters = np.ones(data.shape, dtype=bool)
    passes_filters[:, :, :, :, -1] = 0
    intensities[Features.PASSES_THRESHOLDS] = (Features.AXIS, np.ravel(passes_filters))

    return intensities, labels_with_nan


def test_intensities_to_decoded_image():
    """Test that the decoded image matches the labels object that it is expected to."""

    intensities, labels_with_nan = decoded_intensity_table_factory()

    # test producing a decoded image
    targets_map = TargetsMap(intensities[Features.TARGET].values)
    decoded_image = CombineAdjacentFeatures._intensities_to_decoded_image(
        intensities,
        targets_map,
        mask_filtered_features=False
    )

    # we should be able to map the results back to labels_with_nan using the targets dict
    reverted_labels = targets_map.targets_as_str(np.ravel(decoded_image)).reshape(2, 3, 3)
    assert np.array_equal(reverted_labels, labels_with_nan)


def test_intensities_failing_filters_are_masked_when_requested():
    """
    Test that a masking request causes masked features to be set to zero (background) in the label
    image.
    """

    intensities, labels_with_nan = decoded_intensity_table_factory()

    # test producing a decoded image
    targets_map = TargetsMap(intensities[Features.TARGET].values)
    decoded_image = CombineAdjacentFeatures._intensities_to_decoded_image(
        intensities,
        targets_map,
        mask_filtered_features=True
    )

    # the third column should now be masked
    assert np.all(decoded_image[:, :, -1] == 0)



from starfish.imagestack.imagestack import ImageStack
from starfish.pipeline.algorithmbase import AlgorithmBase


class SegmentationAlgorithmBase(AlgorithmBase):
    def run(self, hybridization_stack: ImageStack, nuclei_stack: ImageStack):
        """Performs registration on the stack provided."""
        raise NotImplementedError()
